# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: virtualbox
short_description: Virtualbox inventory source
description:
  - Get inventory hosts from the local virtualbox installation.
  - Uses a YAML configuration file that ends with virtualbox.(yml|yaml) or vbox.(yml|yaml).
  - The inventory_hostname is always the 'Name' of the virtualbox instance.
  - Groups can be assigned to the VMs using C(VBoxManage). Multiple groups can be assigned by using V(/) as a delimeter.
  - A separate parameter, O(enable_advanced_group_parsing) is exposed to change grouping behaviour. See the parameter documentation
    for details.
extends_documentation_fragment:
  - ansible.builtin.constructed
  - ansible.builtin.inventory_cache
options:
  plugin:
    description: Token that ensures this is a source file for the P(community.general.virtualbox#inventory) plugin.
    type: string
    required: true
    choices: ['virtualbox', 'community.general.virtualbox']
  running_only:
    description: Toggles showing all VMs instead of only those currently running.
    type: boolean
    default: false
  settings_password_file:
    description: Provide a file containing the settings password (equivalent to C(--settingspwfile)).
    type: string
  network_info_path:
    description: Property path to query for network information (C(ansible_host)).
    type: string
    default: "/VirtualBox/GuestInfo/Net/0/V4/IP"
  query:
    description: Create vars from virtualbox properties.
    type: dictionary
    default: {}
  enable_advanced_group_parsing:
    description:
      - The default group parsing rule (when this setting is set to V(false)) is to split the VirtualBox VM's group based
        on the V(/) character and assign the resulting list elements as an Ansible Group.
      - Setting O(enable_advanced_group_parsing=true) changes this behaviour to match VirtualBox's interpretation of groups
        according to U(https://www.virtualbox.org/manual/UserManual.html#gui-vmgroups). Groups are now split using the V(,)
        character, and the V(/) character indicates nested groups.
      - When enabled, a VM that's been configured using V(VBoxManage modifyvm "vm01" --groups "/TestGroup/TestGroup2,/TestGroup3")
        results in the group C(TestGroup2) being a child group of C(TestGroup); and the VM being a part of C(TestGroup2)
        and C(TestGroup3).
    default: false
    type: bool
    version_added: 9.2.0
"""

EXAMPLES = r"""
---
# file must be named vbox.yaml or vbox.yml
plugin: community.general.virtualbox
settings_password_file: /etc/virtualbox/secrets
query:
  logged_in_users: /VirtualBox/GuestInfo/OS/LoggedInUsersList
compose:
  ansible_connection: ('indows' in vbox_Guest_OS)|ternary('winrm', 'ssh')

---
# add hosts (all match with minishift vm) to the group container if any of the vms are in ansible_inventory'
plugin: community.general.virtualbox
groups:
  container: "'minis' in (inventory_hostname)"
"""

import os
from collections.abc import MutableMapping
from subprocess import PIPE, Popen

from ansible.errors import AnsibleParserError
from ansible.module_utils.common.process import get_bin_path
from ansible.module_utils.common.text.converters import to_bytes, to_text
from ansible.plugins.inventory import BaseInventoryPlugin, Cacheable, Constructable

from ansible_collections.community.general.plugins.plugin_utils._unsafe import make_unsafe


class InventoryModule(BaseInventoryPlugin, Constructable, Cacheable):
    """Host inventory parser for ansible using local virtualbox."""

    NAME = "community.general.virtualbox"
    VBOX = "VBoxManage"

    def __init__(self):
        self._vbox_path = None
        super().__init__()

    def _query_vbox_data(self, host, property_path):
        ret = None
        try:
            cmd = [
                self._vbox_path,
                b"guestproperty",
                b"get",
                to_bytes(host, errors="surrogate_or_strict"),
                to_bytes(property_path, errors="surrogate_or_strict"),
            ]
            x = Popen(cmd, stdout=PIPE)
            ipinfo = to_text(x.stdout.read(), errors="surrogate_or_strict")
            if "Value" in ipinfo:
                a, ip = ipinfo.split(":", 1)
                ret = ip.strip()
        except Exception:
            pass
        return ret

    def _set_variables(self, hostvars):
        # set vars in inventory from hostvars
        for host in hostvars:
            query = self.get_option("query")
            # create vars from vbox properties
            if query and isinstance(query, MutableMapping):
                for varname in query:
                    hostvars[host][varname] = self._query_vbox_data(host, query[varname])

            strict = self.get_option("strict")

            # create composite vars
            self._set_composite_vars(self.get_option("compose"), hostvars[host], host, strict=strict)

            # actually update inventory
            for key in hostvars[host]:
                self.inventory.set_variable(host, key, hostvars[host][key])

            # constructed groups based on conditionals
            self._add_host_to_composed_groups(self.get_option("groups"), hostvars[host], host, strict=strict)

            # constructed keyed_groups
            self._add_host_to_keyed_groups(self.get_option("keyed_groups"), hostvars[host], host, strict=strict)

    def _populate_from_cache(self, source_data):
        source_data = make_unsafe(source_data)
        hostvars = source_data.pop("_meta", {}).get("hostvars", {})
        for group in source_data:
            if group == "all":
                continue
            else:
                group = self.inventory.add_group(group)
                hosts = source_data[group].get("hosts", [])
                for host in hosts:
                    self._populate_host_vars([host], hostvars.get(host, {}), group)
                self.inventory.add_child("all", group)
        if not source_data:
            for host in hostvars:
                self.inventory.add_host(host)
                self._populate_host_vars([host], hostvars.get(host, {}))

    def _populate_from_source(self, source_data, using_current_cache=False):
        if using_current_cache:
            self._populate_from_cache(source_data)
            return source_data

        cacheable_results = {"_meta": {"hostvars": {}}}

        hostvars = {}
        prevkey = pref_k = ""
        current_host = None

        # needed to possibly set ansible_host
        netinfo = self.get_option("network_info_path")

        for line in source_data:
            line = to_text(line)
            if ":" not in line:
                continue
            try:
                k, v = line.split(":", 1)
            except Exception:
                # skip non splitable
                continue

            if k.strip() == "":
                # skip empty
                continue

            v = v.strip()
            # found host
            if k.startswith("Name") and "," not in v:  # some setting strings appear in Name
                current_host = make_unsafe(v)
                if current_host not in hostvars:
                    hostvars[current_host] = {}
                    self.inventory.add_host(current_host)

                # try to get network info
                netdata = self._query_vbox_data(current_host, netinfo)
                if netdata:
                    self.inventory.set_variable(current_host, "ansible_host", make_unsafe(netdata))

            # found groups
            elif k == "Groups":
                if self.get_option("enable_advanced_group_parsing"):
                    self._handle_vboxmanage_group_string(v, current_host, cacheable_results)
                else:
                    self._handle_group_string(v, current_host, cacheable_results)
                continue

            else:
                # found vars, accumulate in hostvars for clean inventory set
                pref_k = make_unsafe(f"vbox_{k.strip().replace(' ', '_')}")
                leading_spaces = len(k) - len(k.lstrip(" "))
                if 0 < leading_spaces <= 2:
                    if prevkey not in hostvars[current_host] or not isinstance(hostvars[current_host][prevkey], dict):
                        hostvars[current_host][prevkey] = {}
                    hostvars[current_host][prevkey][pref_k] = make_unsafe(v)
                elif leading_spaces > 2:
                    continue
                else:
                    if v != "":
                        hostvars[current_host][pref_k] = make_unsafe(v)
                if self._ungrouped_host(current_host, cacheable_results):
                    if "ungrouped" not in cacheable_results:
                        cacheable_results["ungrouped"] = {"hosts": []}
                    cacheable_results["ungrouped"]["hosts"].append(current_host)

                prevkey = pref_k

        self._set_variables(hostvars)
        for host in hostvars:
            h = self.inventory.get_host(host)
            cacheable_results["_meta"]["hostvars"][h.name] = h.vars

        return cacheable_results

    def _ungrouped_host(self, host, inventory):
        def find_host(host, inventory):
            for k, v in inventory.items():
                if k == "_meta":
                    continue
                if isinstance(v, dict):
                    yield self._ungrouped_host(host, v)
                elif isinstance(v, list):
                    yield host not in v
            yield True

        return all(find_host(host, inventory))

    def _handle_group_string(self, vboxmanage_group, current_host, cacheable_results):
        """Handles parsing the VM's Group assignment from VBoxManage according to this inventory's initial implementation."""
        # The original implementation of this inventory plugin treated `/` as
        # a delimeter to split and use as Ansible Groups.
        for group in vboxmanage_group.split("/"):
            if group:
                group = make_unsafe(group)
                group = self.inventory.add_group(group)
                self.inventory.add_child(group, current_host)
                if group not in cacheable_results:
                    cacheable_results[group] = {"hosts": []}
                cacheable_results[group]["hosts"].append(current_host)

    def _handle_vboxmanage_group_string(self, vboxmanage_group, current_host, cacheable_results):
        """Handles parsing the VM's Group assignment from VBoxManage according to VirtualBox documentation."""
        # Per the VirtualBox documentation, a VM can be part of many groups,
        # and it is possible to have nested groups.
        # Many groups are separated by commas ",", and nested groups use
        # slash "/".
        # https://www.virtualbox.org/manual/UserManual.html#gui-vmgroups
        # Multi groups: VBoxManage modifyvm "vm01" --groups "/TestGroup,/TestGroup2"
        # Nested groups: VBoxManage modifyvm "vm01" --groups "/TestGroup/TestGroup2"

        for group in vboxmanage_group.split(","):
            if not group:
                # We could get an empty element due how to split works, and
                # possible assignments from VirtualBox. e.g. ,/Group1
                continue

            if group == "/":
                # This is the "root" group. We get here if the VM was not
                # assigned to a particular group. Consider the host to be
                # unassigned to a group.
                continue

            parent_group = None
            for subgroup in group.split("/"):
                if not subgroup:
                    # Similarly to above, we could get an empty element.
                    # e.g //Group1
                    continue

                if subgroup == "/":
                    # "root" group.
                    # Consider the host to be unassigned
                    continue

                subgroup = make_unsafe(subgroup)
                subgroup = self.inventory.add_group(subgroup)
                if parent_group is not None:
                    self.inventory.add_child(parent_group, subgroup)
                self.inventory.add_child(subgroup, current_host)
                if subgroup not in cacheable_results:
                    cacheable_results[subgroup] = {"hosts": []}
                cacheable_results[subgroup]["hosts"].append(current_host)

                parent_group = subgroup

    def verify_file(self, path):
        valid = False
        if super().verify_file(path):
            if path.endswith(("virtualbox.yaml", "virtualbox.yml", "vbox.yaml", "vbox.yml")):
                valid = True
        return valid

    def parse(self, inventory, loader, path, cache=True):
        try:
            self._vbox_path = get_bin_path(self.VBOX)
        except ValueError as e:
            raise AnsibleParserError(e) from e

        super().parse(inventory, loader, path)

        cache_key = self.get_cache_key(path)

        config_data = self._read_config_data(path)

        # set _options from config data
        self._consume_options(config_data)

        source_data = None
        if cache:
            cache = self.get_option("cache")

        update_cache = False
        if cache:
            try:
                source_data = self._cache[cache_key]
            except KeyError:
                update_cache = True

        if not source_data:
            b_pwfile = to_bytes(
                self.get_option("settings_password_file"), errors="surrogate_or_strict", nonstring="passthru"
            )
            running = self.get_option("running_only")

            # start getting data
            cmd = [self._vbox_path, b"list", b"-l"]
            if running:
                cmd.append(b"runningvms")
            else:
                cmd.append(b"vms")

            if b_pwfile and os.path.exists(b_pwfile):
                cmd.append(b"--settingspwfile")
                cmd.append(b_pwfile)

            try:
                p = Popen(cmd, stdout=PIPE)
            except Exception as e:
                raise AnsibleParserError(str(e)) from e

            source_data = p.stdout.read().splitlines()

        using_current_cache = cache and not update_cache
        cacheable_results = self._populate_from_source(source_data, using_current_cache)

        if update_cache:
            self._cache[cache_key] = cacheable_results



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Andrew Klychkov (@Andersson007) <andrew.a.klychkov@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)


from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = r'''
---
module: mysql_info
short_description: Gather information about MySQL or MariaDB servers
description:
- Gathers information about MySQL or MariaDB servers.

options:
  filter:
    description:
    - Limit the collected information by comma separated string or YAML list.
    - Allowable values are C(version), C(databases), C(settings), C(global_status),
      C(users), C(users_info), C(engines), C(master_status), C(slave_status), C(slave_hosts).
    - By default, collects all subsets.
    - You can use '!' before value (for example, C(!settings)) to exclude it from the information.
    - If you pass including and excluding values to the filter, for example, I(filter=!settings,version),
      the excluding values, C(!settings) in this case, will be ignored.
    type: list
    elements: str
  login_db:
    description:
    - Database name to connect to.
    - It makes sense if I(login_user) is allowed to connect to a specific database only.
    type: str
  exclude_fields:
    description:
    - List of fields which are not needed to collect.
    - "Supports elements: C(db_size), C(db_table_count). Unsupported elements will be ignored."
    type: list
    elements: str
    version_added: '0.1.0'
  return_empty_dbs:
    description:
    - Includes names of empty databases to returned dictionary.
    type: bool
    default: false

notes:
- Compatible with MariaDB or MySQL.
- Calculating the size of a database might be slow, depending on the number and size of tables in it.
  To avoid this, use I(exclude_fields=db_size).
- filters C(users_info) doesn't support MariaDB roles.

attributes:
  check_mode:
    support: full
  idempotent:
    support: full

seealso:
- module: ansible.mysql.mysql_variables
- module: ansible.mysql.mysql_db
- module: ansible.mysql.mysql_user
- module: ansible.mysql.mysql_replication

author:
- Andrew Klychkov (@Andersson007)
- Sebastian Gumprich (@rndmh3ro)
- Laurent Indermühle (@laurent-indermuehle)

extends_documentation_fragment:
- ansible.mysql.mysql
'''

EXAMPLES = r'''
# Display info from mysql-hosts group (using creds from ~/.my.cnf to connect):
# ansible mysql-hosts -m mysql_info

# Display only databases and users info:
# ansible mysql-hosts -m mysql_info -a 'filter=databases,users'

# Display all users privileges:
# ansible mysql-hosts -m mysql_info -a 'filter=users_info'

# Display only slave status:
# ansible standby -m mysql_info -a 'filter=slave_status'

# Display all info from databases group except settings:
# ansible databases -m mysql_info -a 'filter=!settings'

# If you encounter the "Please explicitly state intended protocol" error,
# use the login_unix_socket argument
- name: Collect all possible information using passwordless root access
  ansible.mysql.mysql_info:
    login_user: root
    login_unix_socket: /run/mysqld/mysqld.sock

- name: Get MySQL version with non-default credentials
  ansible.mysql.mysql_info:
    login_user: mysuperuser
    login_password: mysuperpass
    filter: version

- name: Collect all info except settings and users by root
  ansible.mysql.mysql_info:
    login_user: root
    login_password: rootpass
    filter: "!settings,!users"

- name: Collect info about databases and version using ~/.my.cnf as a credential file
  become: true
  ansible.mysql.mysql_info:
    filter:
      - databases
      - version

- name: Collect info about databases and version using ~alice/.my.cnf as a credential file
  become: true
  ansible.mysql.mysql_info:
    config_file: /home/alice/.my.cnf
    filter:
      - databases
      - version

- name: Collect info about databases including empty and excluding their sizes
  become: true
  ansible.mysql.mysql_info:
    config_file: /home/alice/.my.cnf
    filter:
      - databases
    exclude_fields: db_size
    return_empty_dbs: true

- name: Clone users from one server to another
  block:
    # Step 1
    - name: Fetch information from a source server
      delegate_to: server_source
      ansible.mysql.mysql_info:
        filter:
          - users_info
      register: result

    # Step 2
    # Don't work with sha256_password and cache_sha2_password
    - name: Clone users fetched in a previous task to a target server
      ansible.mysql.mysql_user:
        name: "{{ item.name }}"
        host: "{{ item.host }}"
        plugin: "{{ item.plugin | default(omit) }}"
        plugin_hash_string: "{{ item.plugin_hash_string | default(omit) }}"
        tls_requires: "{{ item.tls_requires | default(omit) }}"
        priv: "{{ item.priv | default(omit) }}"
        resource_limits: "{{ item.resource_limits | default(omit) }}"
        locked: "{{ item.locked | default(omit) }}"
        column_case_sensitive: true
        state: present
      loop: "{{ result.users_info }}"
      loop_control:
        label: "{{ item.name }}@{{ item.host }}"
      when:
        - item.name != 'root'  # In case you don't want to import admin accounts
        - item.name != 'mariadb.sys'
        - item.name != 'mysql'
        - item.name != 'PUBLIC'  # MariaDB roles are not supported
'''

RETURN = r'''
server_engine:
  description: Database server engine.
  returned: if not excluded by filter
  type: str
  sample: 'MariaDB'
  version_added: '3.10.0'
version:
  description: Database server version.
  returned: if not excluded by filter
  type: dict
  sample: { "version": { "major": 5, "minor": 5, "release": 60, "suffix": "MariaDB", "full": "5.5.60-MariaDB" } }
  contains:
    major:
      description: Major server version.
      returned: if not excluded by filter
      type: int
      sample: 5
    minor:
      description: Minor server version.
      returned: if not excluded by filter
      type: int
      sample: 5
    release:
      description: Release server version.
      returned: if not excluded by filter
      type: int
      sample: 60
    suffix:
      description: Server suffix, for example MySQL, MariaDB, other or none.
      returned: if not excluded by filter
      type: str
      sample: "MariaDB"
    full:
      description: Full server version.
      returned: if not excluded by filter
      type: str
      sample: "5.5.60-MariaDB"
databases:
  description: Information about databases.
  returned: if not excluded by filter
  type: dict
  sample:
  - { "mysql": { "size": 656594, "tables": 31 }, "information_schema": { "size": 73728, "tables": 79 } }
  contains:
    size:
      description: Database size in bytes.
      returned: if not excluded by filter
      type: dict
      sample: { 'size': 656594 }
    tables:
      description: Count of tables and views in that database.
      returned: if not excluded by filter
      type: dict
      sample: { 'tables': 12 }
      version_added: '3.11.0'
settings:
  description: Global settings (variables) information.
  returned: if not excluded by filter
  type: dict
  sample:
  - { "innodb_open_files": 300, innodb_page_size": 16384 }
global_status:
  description: Global status information.
  returned: if not excluded by filter
  type: dict
  sample:
  - { "Innodb_buffer_pool_read_requests": 123, "Innodb_buffer_pool_reads": 32 }
users:
  description: Return a dictionnary of users grouped by host and with global privileges only.
  returned: if not excluded by filter
  type: dict
  sample:
  - { "localhost": { "root": { "Alter_priv": "Y", "Alter_routine_priv": "Y" } } }
users_info:
  description:
    - Information about users accounts.
    - The output can be used as an input of the M(ansible.mysql.mysql_user) plugin.
    - Useful when migrating accounts to another server or to create an inventory.
    - Does not support proxy privileges. If an account has proxy privileges, they won't appear in the output.
    - Causes issues with authentications plugins C(sha256_password) and C(caching_sha2_password).
      If the output is fed to M(ansible.mysql.mysql_user), the
      ``plugin_hash_string`` will most likely be unreadable due to non-binary
      characters.
    - The "locked" field was aded in ``ansible.mysql`` 3.13.
  returned: if not excluded by filter
  type: dict
  sample:
  - { "plugin_hash_string": '*1234567',
      "name": "user1",
      "host": "host.com",
      "plugin": "mysql_native_password",
      "priv": "db1.*:SELECT/db2.*:SELECT",
      "resource_limits": { "MAX_USER_CONNECTIONS": 100 },
      "tls_requires": { "SSL": null },
      "locked": false }
  version_added: '3.8.0'
engines:
  description: Information about the server's storage engines.
  returned: if not excluded by filter
  type: dict
  sample:
  - { "CSV": { "Comment": "CSV storage engine", "Savepoints": "NO", "Support": "YES", "Transactions": "NO", "XA": "NO" } }
master_status:
  description: Master status information.
  returned: if master
  type: dict
  sample:
  - { "Binlog_Do_DB": "", "Binlog_Ignore_DB": "mysql", "File": "mysql-bin.000001", "Position": 769 }
slave_status:
  description: Slave status information.
  returned: if standby
  type: dict
  sample:
  - { "192.168.1.101": { "3306": { "replication_user": { "Connect_Retry": 60, "Exec_Master_Log_Pos": 769,  "Last_Errno": 0 } } } }
slave_hosts:
  description: Slave status information.
  returned: if master
  type: dict
  sample:
  - { "2": { "Host": "", "Master_id": 1, "Port": 3306 } }
connector_name:
  description: Name of the python connector used by the module. When the connector is not identified, returns C(Unknown).
  returned: always
  type: str
  sample:
  - "pymysql"
  version_added: '3.6.0'
connector_version:
  description: Version of the python connector used by the module. When the connector is not identified, returns C(Unknown).
  returned: always
  type: str
  sample:
  - "1.0.2"
  version_added: '3.6.0'
'''

from decimal import Decimal

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.ansible.mysql.plugins.module_utils.command_resolver import (
    CommandResolver
)
from ansible_collections.ansible.mysql.plugins.module_utils.mysql import (
    mysql_connect,
    mysql_common_argument_spec,
    mysql_driver,
    mysql_driver_fail_msg,
    get_connector_name,
    get_connector_version,
    get_server_implementation,
    get_server_version,
)

from ansible_collections.ansible.mysql.plugins.module_utils.user import (
    privileges_get,
    get_resource_limits,
    get_existing_authentication,
    get_user_implementation,
    user_is_locked,
)
from ansible.module_utils.common.text.converters import to_native


# ===========================================
# MySQL module specific support methods.
#

class MySQL_Info(object):

    """Class for collection MySQL instance information.

    Arguments:
        module (AnsibleModule): Object of AnsibleModule class.
        cursor (pymysql/mysql-python): Cursor class for interaction with
            the database.

    Note:
        If you need to add a new subset:
        1. add a new key with the same name to self.info attr in self.__init__()
        2. add a new private method to get the information
        3. add invocation of the new method to self.__collect()
        4. add info about the new subset to the DOCUMENTATION block
        5. add info about the new subset with an example to RETURN block
    """

    def __init__(self, module, cursor, server_implementation, server_version, user_implementation):
        self.module = module
        self.cursor = cursor
        self.server_implementation = server_implementation
        self.server_version = server_version
        self.user_implementation = user_implementation
        self.command_resolver = CommandResolver(self.server_implementation, self.server_version)
        self.info = {
            'version': {},
            'databases': {},
            'settings': {},
            'global_status': {},
            'engines': {},
            'users': {},
            'users_info': {},
            'master_status': {},
            'slave_hosts': {},
            'slave_status': {},
        }

    def get_info(self, filter_, exclude_fields, return_empty_dbs):
        """Get MySQL instance information based on filter_.

        Arguments:
            filter_ (list): List of collected subsets (e.g., databases, users, etc.),
                when it is empty, return all available information.
        """

        inc_list = []
        exc_list = []

        if filter_:
            partial_info = {}

            for fi in filter_:
                if fi.lstrip('!') not in self.info:
                    self.module.warn('filter element: %s is not allowable, ignored' % fi)
                    continue

                if fi[0] == '!':
                    exc_list.append(fi.lstrip('!'))

                else:
                    inc_list.append(fi)

            if inc_list:
                self.__collect(exclude_fields, return_empty_dbs, set(inc_list))

                for i in self.info:
                    if i in inc_list:
                        partial_info[i] = self.info[i]

            else:
                not_in_exc_list = list(set(self.info) - set(exc_list))
                self.__collect(exclude_fields, return_empty_dbs, set(not_in_exc_list))

                for i in self.info:
                    if i not in exc_list:
                        partial_info[i] = self.info[i]

            return partial_info

        else:
            self.__collect(exclude_fields, return_empty_dbs, set(self.info))
            return self.info

    def __collect(self, exclude_fields, return_empty_dbs, wanted):
        """Collect all possible subsets."""
        if 'version' in wanted or 'settings' in wanted:
            self.__get_global_variables()

        if 'databases' in wanted:
            self.__get_databases(exclude_fields, return_empty_dbs)

        if 'global_status' in wanted:
            self.__get_global_status()

        if 'engines' in wanted:
            self.__get_engines()

        if 'users' in wanted:
            self.__get_users()

        if 'users_info' in wanted:
            self.__get_users_info()

        if 'master_status' in wanted:
            self.__get_master_status()

        if 'slave_status' in wanted:
            self.__get_slave_status()

        if 'slave_hosts' in wanted:
            self.__get_slaves()

    def __get_engines(self):
        """Get storage engines info."""
        res = self.__exec_sql('SHOW ENGINES')

        if res:
            for line in res:
                engine = line['Engine']
                self.info['engines'][engine] = {}

                for vname, val in line.items():
                    if vname != 'Engine':
                        self.info['engines'][engine][vname] = val

    def __convert(self, val):
        """Convert unserializable data."""
        try:
            if isinstance(val, Decimal):
                val = float(val)
            else:
                val = int(val)

        except ValueError:
            pass

        except TypeError:
            pass

        return val

    def __get_global_variables(self):
        """Get global variables (instance settings)."""
        res = self.__exec_sql('SHOW GLOBAL VARIABLES')

        if res:
            for var in res:
                self.info['settings'][var['Variable_name']] = self.__convert(var['Value'])

            # version = ["5", "5," "60-MariaDB]
            version = self.info['settings']['version'].split('.')

            # full_version = "5.5.60-MariaDB"
            full = self.info['settings']['version']

            # release = "60"
            release = version[2].split('-')[0]

            # check if a suffix exists by counting the length
            if len(version[2].split('-')) > 1:
                # suffix = "MariaDB"
                suffix = version[2].split('-', 1)[1]
            else:
                suffix = ""

            self.info['version'] = dict(
                # major = "5"
                major=int(version[0]),
                # minor = "5"
                minor=int(version[1]),
                release=int(release),
                suffix=str(suffix),
                full=str(full),
            )

    def __get_global_status(self):
        """Get global status."""
        res = self.__exec_sql('SHOW GLOBAL STATUS')

        if res:
            for var in res:
                self.info['global_status'][var['Variable_name']] = self.__convert(var['Value'])

    def __get_master_status(self):
        """Get master status if the instance is a master."""
        query = self.command_resolver.resolve_command("SHOW MASTER STATUS")
        res = self.__exec_sql(query)
        if res:
            for line in res:
                for vname, val in line.items():
                    self.info['master_status'][vname] = self.__convert(val)

    def __get_slave_status(self):
        """Get slave status if the instance is a slave."""
        query = self.command_resolver.resolve_command("SHOW SLAVE STATUS")
        res = self.__exec_sql(query)
        if res:
            for line in res:
                host = line.get('Master_Host') or line.get('Source_Host')
                if host not in self.info['slave_status']:
                    self.info['slave_status'][host] = {}

                port = line.get('Master_Port') or line.get('Source_Port')
                if port not in self.info['slave_status'][host]:
                    self.info['slave_status'][host][port] = {}

                user = line.get('Master_User') or line.get('Source_User')
                if user not in self.info['slave_status'][host][port]:
                    self.info['slave_status'][host][port][user] = {}

                for vname, val in line.items():
                    if vname not in ('Master_Host', 'Master_Port', 'Master_User', 'Source_Host', 'Source_Port', 'Source_User'):
                        self.info['slave_status'][host][port][user][vname] = self.__convert(val)

    def __get_slaves(self):
        """Get slave hosts info if the instance is a master."""
        query = self.command_resolver.resolve_command("SHOW SLAVE HOSTS")
        res = self.__exec_sql(query)
        if res:
            for line in res:
                srv_id = line['Server_id']
                if srv_id not in self.info['slave_hosts']:
                    self.info['slave_hosts'][srv_id] = {}

                for vname, val in line.items():
                    if vname != 'Server_id':
                        self.info['slave_hosts'][srv_id][vname] = self.__convert(val)

    def __get_users(self):
        """Get user info."""
        res = self.__exec_sql('SELECT * FROM mysql.user')
        if res:
            for line in res:
                host = line['Host']
                if host not in self.info['users']:
                    self.info['users'][host] = {}

                user = line['User']
                self.info['users'][host][user] = {}

                for vname, val in line.items():
                    if vname not in ('Host', 'User'):
                        self.info['users'][host][user][vname] = self.__convert(val)

    def __get_users_info(self):
        """Get user privileges, passwords, resources_limits, ...

        Query the server to get all the users and return a string
        of privileges that can be used by the mysql_user plugin.
        For instance:

        "users_info": [
            {
                "host": "users_info.com",
                "priv": "*.*: ALL,GRANT",
                "name": "users_info_adm"
            },
            {
                "host": "users_info.com",
                "priv": "`mysql`.*: SELECT/`users_info_db`.*: SELECT",
                "name": "users_info_multi"
            }
        ]
        """
        res = self.__exec_sql('SELECT * FROM mysql.user')
        if not res:
            return None

        output = list()
        for line in res:
            user = line['User']
            host = line['Host']

            # MariaDB roles have no host
            is_role = self.server_implementation == 'mariadb' and not host
            user_priv = privileges_get(self.cursor, user, host, maria_role=is_role)

            if not user_priv:
                self.module.warn("No privileges found for %s on host %s" % (user, host))
                continue

            priv_string = list()
            for db_table, priv in user_priv.items():
                # Proxy privileges are hard to work with because of different quotes or
                # backticks like ''@'', ''@'%' or even ``@``. In addition, MySQL will
                # forbid you to grant a proxy privileges through TCP.
                if set(priv) == {'PROXY', 'GRANT'} or set(priv) == {'PROXY'}:
                    continue

                unquote_db_table = db_table.replace('`', '').replace("'", '')
                priv_string.append('%s:%s' % (unquote_db_table, ','.join(priv)))

            # Only keep *.* USAGE if it's the only user privilege given
            if len(priv_string) > 1 and '*.*:USAGE' in priv_string:
                priv_string.remove('*.*:USAGE')

            resource_limits = get_resource_limits(self.cursor, user, host)
            copy_ressource_limits = dict.copy(resource_limits)

            tls_requires = self.user_implementation.get_tls_requires(
                self.cursor, user, host)

            output_dict = {
                'name': user,
                'host': host,
                'priv': '/'.join(priv_string),
                'resource_limits': copy_ressource_limits,
                'tls_requires': tls_requires,
            }

            # Prevent returning a resource limit if empty
            if resource_limits:
                for key, value in resource_limits.items():
                    if value == 0:
                        del output_dict['resource_limits'][key]
                if len(output_dict['resource_limits']) == 0:
                    del output_dict['resource_limits']

            # Prevent returning tls_require if empty
            if not tls_requires:
                del output_dict['tls_requires']

            authentications = get_existing_authentication(self.cursor, user, host)
            if authentications:
                output_dict.update(authentications[0])
                # https://github.com/ansible-collections/ansible.mysql/pull/629
                output_dict.pop('plugin_auth_string', None)

            if line.get('is_role') and line['is_role'] == 'N':
                output_dict['locked'] = user_is_locked(self.cursor, user, host)

            # TODO password_option
            # but both are not supported by mysql_user atm. So no point yet.

            output.append(output_dict)

        self.info['users_info'] = output

    def __get_databases(self, exclude_fields, return_empty_dbs):
        """Get info about databases."""

        def is_field_included(field_name):
            return not exclude_fields or 'db_{}'.format(field_name) not in exclude_fields

        def create_db_info(db_data):
            info = {}
            if is_field_included('size'):
                info['size'] = int(db_data.get('size', 0) or 0)
            if is_field_included('table_count'):
                info['tables'] = int(db_data.get('tables', 0) or 0)
            return info

        # Build the main query
        query_parts = ['SELECT table_schema AS "name"']
        if is_field_included('size'):
            query_parts.append('SUM(data_length + index_length) AS "size"')
        if is_field_included('table_count'):
            query_parts.append('COUNT(table_name) as "tables"')

        query = "{} FROM information_schema.TABLES GROUP BY table_schema".format(", ".join(query_parts))

        # Get and process databases with tables
        databases = self.__exec_sql(query) or []
        for db in databases:
            self.info['databases'][db['name']] = create_db_info(db)

        # Handle empty databases if requested
        if return_empty_dbs:
            empty_databases = self.__exec_sql('SHOW DATABASES') or []
            for db in empty_databases:
                db_name = db['Database']
                if db_name not in self.info['databases']:
                    self.info['databases'][db_name] = create_db_info({})

    def __exec_sql(self, query, ddl=False):
        """Execute SQL.

        Arguments:
            ddl (bool): If True, return True or False.
                Used for queries that don't return any rows
                (mainly for DDL queries) (default False).
        """
        try:
            self.cursor.execute(query)

            if not ddl:
                res = self.cursor.fetchall()
                return res
            return True

        except Exception as e:
            self.module.fail_json(msg="Cannot execute SQL '%s': %s" % (query, to_native(e)))
        return False


# ===========================================
# Module execution.
#


def main():
    argument_spec = mysql_common_argument_spec()
    argument_spec.update(
        login_db=dict(type='str'),
        filter=dict(type='list', elements='str'),
        exclude_fields=dict(type='list', elements='str'),
        return_empty_dbs=dict(type='bool', default=False),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    db = module.params['login_db']
    connect_timeout = module.params['connect_timeout']
    login_user = module.params['login_user']
    login_password = module.params['login_password']
    ssl_cert = module.params['client_cert']
    ssl_key = module.params['client_key']
    ssl_ca = module.params['ca_cert']
    check_hostname = module.params['check_hostname']
    config_file = module.params['config_file']
    filter_ = module.params['filter']
    exclude_fields = module.params['exclude_fields']
    return_empty_dbs = module.params['return_empty_dbs']

    if filter_:
        filter_ = [f.strip() for f in filter_]

    if exclude_fields:
        exclude_fields = set([f.strip() for f in exclude_fields])

    if mysql_driver is None:
        module.fail_json(msg=mysql_driver_fail_msg)

    connector_name = get_connector_name(mysql_driver)
    connector_version = get_connector_version(mysql_driver)

    try:
        cursor, db_conn = mysql_connect(module, login_user, login_password,
                                        config_file, ssl_cert, ssl_key, ssl_ca, db,
                                        check_hostname=check_hostname,
                                        connect_timeout=connect_timeout, cursor_class='DictCursor')
    except Exception as e:
        msg = ('unable to connect to database using %s %s, check login_user '
               'and login_password are correct or %s has the credentials. '
               'Exception message: %s' % (connector_name, connector_version, config_file, to_native(e)))
        module.fail_json(msg)

    server_implementation = get_server_implementation(cursor)
    server_version = get_server_version(cursor)
    user_implementation = get_user_implementation(cursor)

    ###############################
    # Create object and do main job

    mysql = MySQL_Info(module, cursor, server_implementation, server_version, user_implementation)

    module.exit_json(changed=False,
                     server_engine='MariaDB' if server_implementation == 'mariadb' else 'MySQL',
                     connector_name=connector_name,
                     connector_version=connector_version,
                     **mysql.get_info(filter_, exclude_fields, return_empty_dbs))


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright Ansible Team
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: github_release
short_description: Interact with GitHub Releases
description:
  - Fetch metadata about GitHub Releases.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  token:
    description:
      - GitHub Personal Access Token for authenticating. Mutually exclusive with O(password).
    type: str
  user:
    description:
      - The GitHub account that owns the repository.
    type: str
    required: true
  password:
    description:
      - The GitHub account password for the user. Mutually exclusive with O(token).
    type: str
  repo:
    description:
      - Repository name.
    type: str
    required: true
  action:
    description:
      - Action to perform.
    type: str
    required: true
    choices: ['latest_release', 'create_release']
  tag:
    description:
      - Tag name when creating a release. Required when using O(action=create_release).
    type: str
  target:
    description:
      - Target of release when creating a release.
    type: str
  name:
    description:
      - Name of release when creating a release.
    type: str
  body:
    description:
      - Description of the release when creating a release.
    type: str
  draft:
    description:
      - Sets if the release is a draft or not. (boolean).
    type: bool
    default: false
  prerelease:
    description:
      - Sets if the release is a prerelease or not. (boolean).
    type: bool
    default: false

author:
  - "Adrian Moisey (@adrianmoisey)"
requirements:
  - "github3.py >= 1.0.0a3"
"""

EXAMPLES = r"""
- name: Get latest release of a public repository
  community.general.github_release:
    user: ansible
    repo: ansible
    action: latest_release

- name: Get latest release of testuseer/testrepo
  community.general.github_release:
    token: tokenabc1234567890
    user: testuser
    repo: testrepo
    action: latest_release

- name: Get latest release of test repo using username and password
  community.general.github_release:
    user: testuser
    password: secret123
    repo: testrepo
    action: latest_release

- name: Create a new release
  community.general.github_release:
    token: tokenabc1234567890
    user: testuser
    repo: testrepo
    action: create_release
    tag: test
    target: master
    name: My Release
    body: Some description
"""

RETURN = r"""
tag:
  description: Version of the created/latest release.
  type: str
  returned: success
  sample: 1.1.0
"""

import traceback

GITHUB_IMP_ERR = None
try:
    import github3

    HAS_GITHUB_API = True
except ImportError:
    GITHUB_IMP_ERR = traceback.format_exc()
    HAS_GITHUB_API = False

from ansible.module_utils.basic import AnsibleModule, missing_required_lib


def main():
    module = AnsibleModule(
        argument_spec=dict(
            repo=dict(required=True),
            user=dict(required=True),
            password=dict(no_log=True),
            token=dict(no_log=True),
            action=dict(required=True, choices=["latest_release", "create_release"]),
            tag=dict(type="str"),
            target=dict(type="str"),
            name=dict(type="str"),
            body=dict(type="str"),
            draft=dict(type="bool", default=False),
            prerelease=dict(type="bool", default=False),
        ),
        supports_check_mode=True,
        mutually_exclusive=(("password", "token"),),
        required_if=[("action", "create_release", ["tag"]), ("action", "create_release", ["password", "token"], True)],
    )

    if not HAS_GITHUB_API:
        module.fail_json(msg=missing_required_lib("github3.py >= 1.0.0a3"), exception=GITHUB_IMP_ERR)

    repo = module.params["repo"]
    user = module.params["user"]
    password = module.params["password"]
    login_token = module.params["token"]
    action = module.params["action"]
    tag = module.params.get("tag")
    target = module.params.get("target")
    name = module.params.get("name")
    body = module.params.get("body")
    draft = module.params.get("draft")
    prerelease = module.params.get("prerelease")

    # login to github
    try:
        if password:
            gh_obj = github3.login(user, password=password)
        elif login_token:
            gh_obj = github3.login(token=login_token)
        else:
            gh_obj = github3.GitHub()

        # GitHub's token formats:
        #   - ghp_          - Personal access token (classic)
        #   - github_pat_   - Fine-grained personal access token
        #   - gho_          - OAuth access token
        #   - ghu_          - User access token for a GitHub App
        #   - ghs_          - Installation access token for a GitHub App
        #   - ghr_          - Refresh token for a GitHub App
        #
        # References:
        #   https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/about-authentication-to-github#githubs-token-formats
        #
        # Test if we're actually logged in, but skip this check for some token prefixes
        SKIPPED_TOKEN_PREFIXES = ["ghs_"]
        if password or (login_token and not any(login_token.startswith(prefix) for prefix in SKIPPED_TOKEN_PREFIXES)):
            gh_obj.me()
    except github3.exceptions.AuthenticationFailed as e:
        module.fail_json(
            msg=f"Failed to connect to GitHub: {e}",
            details=f"Please check username and password or token for repository {repo}",
        )
    except github3.exceptions.GitHubError as e:
        module.fail_json(
            msg=f"GitHub API error: {e}", details=f"Please check username and password or token for repository {repo}"
        )

    repository = gh_obj.repository(user, repo)

    if not repository:
        module.fail_json(msg=f"Repository {user}/{repo} doesn't exist")

    if action == "latest_release":
        release = repository.latest_release()
        if release:
            module.exit_json(tag=release.tag_name)
        else:
            module.exit_json(tag=None)

    if action == "create_release":
        release_exists = repository.release_from_tag(tag)
        if release_exists:
            module.exit_json(changed=False, msg=f"Release for tag {tag} already exists.")

        release = repository.create_release(tag, target, name, body, draft, prerelease)
        if release:
            module.exit_json(changed=True, tag=release.tag_name)
        else:
            module.exit_json(changed=False, tag=None)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright 2016 Red Hat | Ansible
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_prune

short_description: Allows to prune various docker objects

description:
  - Allows to run C(docker container prune), C(docker image prune), C(docker network prune) and C(docker volume prune) through
    the Docker API.
extends_documentation_fragment:
  - community.docker._docker.api_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker

attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
  idempotent:
    support: full

options:
  containers:
    description:
      - Whether to prune containers.
    type: bool
    default: false
  containers_filters:
    description:
      - A dictionary of filter values used for selecting containers to delete.
      - 'For example, C(until: 24h).'
      - See L(the docker documentation,https://docs.docker.com/engine/reference/commandline/container_prune/#filtering) for
        more information on possible filters.
    type: dict
  images:
    description:
      - Whether to prune images.
    type: bool
    default: false
  images_filters:
    description:
      - A dictionary of filter values used for selecting images to delete.
      - 'For example, C(dangling: true).'
      - See L(the docker documentation,https://docs.docker.com/engine/reference/commandline/image_prune/#filtering) for more
        information on possible filters.
    type: dict
  networks:
    description:
      - Whether to prune networks.
    type: bool
    default: false
  networks_filters:
    description:
      - A dictionary of filter values used for selecting networks to delete.
      - See L(the docker documentation,https://docs.docker.com/engine/reference/commandline/network_prune/#filtering) for
        more information on possible filters.
    type: dict
  volumes:
    description:
      - Whether to prune volumes.
    type: bool
    default: false
  volumes_filters:
    description:
      - A dictionary of filter values used for selecting volumes to delete.
      - See L(the docker documentation,https://docs.docker.com/engine/reference/commandline/volume_prune/#filtering) for more
        information on possible filters.
    type: dict
  builder_cache:
    description:
      - Whether to prune the builder cache.
    type: bool
    default: false
  builder_cache_all:
    description:
      - Whether to remove all types of build cache.
    type: bool
    default: false
    version_added: 3.10.0
  builder_cache_filters:
    description:
      - A dictionary of filter values used for selecting images to delete.
      - 'For example, C(until: 10m).'
      - See L(the API documentation,https://docs.docker.com/engine/api/v1.44/#tag/Image/operation/BuildPrune) for more information
        on possible filters.
    type: dict
    version_added: 3.10.0
  builder_cache_keep_storage:
    description:
      - Amount of disk space to keep for cache in format C(<number>[<unit>]).".
      - Number is a positive integer. Unit can be one of V(B) (byte), V(K) (kibibyte, 1024B), V(M) (mebibyte), V(G) (gibibyte),
        V(T) (tebibyte), or V(P) (pebibyte).
      - Omitting the unit defaults to bytes.
    type: str
    version_added: 3.10.0

author:
  - "Felix Fontein (@felixfontein)"

notes:
  - The module always returned C(changed=false) before community.docker 3.5.1.
requirements:
  - "Docker API >= 1.25"
"""

EXAMPLES = r"""
---
- name: Prune containers older than 24h
  community.docker.docker_prune:
    containers: true
    containers_filters:
      # only consider containers created more than 24 hours ago
      until: 24h

- name: Prune containers with labels
  community.docker.docker_prune:
    containers: true
    containers_filters:
      # Prune containers whose "foo" label has value "bar", and
      # whose "bam" label has value "baz". If you only want to
      # compare one label, you can provide it as a string instead
      # of a list with one element.
      label:
        - foo=bar
        - bam=baz
      # Prune containers whose label "bar" does *not* have value
      # "baz". If you want to avoid more than one label, you can
      # provide a list of multiple label-value pairs.
      "label!": bar=baz

- name: Prune everything
  community.docker.docker_prune:
    containers: true
    images: true
    networks: true
    volumes: true
    builder_cache: true

- name: Prune everything (including non-dangling images)
  community.docker.docker_prune:
    containers: true
    images: true
    images_filters:
      dangling: false
    networks: true
    volumes: true
    builder_cache: true
"""

RETURN = r"""
# containers
containers:
  description:
    - List of IDs of deleted containers.
  returned: O(containers=true)
  type: list
  elements: str
  sample: []
containers_space_reclaimed:
  description:
    - Amount of reclaimed disk space from container pruning in bytes.
  returned: O(containers=true)
  type: int
  sample: 0

# images
images:
  description:
    - List of IDs of deleted images.
  returned: O(images=true)
  type: list
  elements: str
  sample: []
images_space_reclaimed:
  description:
    - Amount of reclaimed disk space from image pruning in bytes.
  returned: O(images=true)
  type: int
  sample: 0

# networks
networks:
  description:
    - List of IDs of deleted networks.
  returned: O(networks=true)
  type: list
  elements: str
  sample: []

# volumes
volumes:
  description:
    - List of IDs of deleted volumes.
  returned: O(volumes=true)
  type: list
  elements: str
  sample: []
volumes_space_reclaimed:
  description:
    - Amount of reclaimed disk space from volumes pruning in bytes.
  returned: O(volumes=true)
  type: int
  sample: 0

# builder_cache
builder_cache_space_reclaimed:
  description:
    - Amount of reclaimed disk space from builder cache pruning in bytes.
  returned: O(builder_cache=true)
  type: int
  sample: 0
builder_cache_caches_deleted:
  description:
    - The build caches that were deleted.
  returned: O(builder_cache=true) and API version is 1.39 or later
  type: list
  elements: str
  sample: []
  version_added: 3.10.0
"""

import traceback

from ansible.module_utils.common.text.formatters import human_to_bytes

from ansible_collections.community.docker.plugins.module_utils._api.errors import (
    DockerException,
)
from ansible_collections.community.docker.plugins.module_utils._api.utils.utils import (
    convert_filters,
)
from ansible_collections.community.docker.plugins.module_utils._common_api import (
    AnsibleDockerClient,
    RequestException,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    clean_dict_booleans_for_docker_api,
)


def main() -> None:
    argument_spec = {
        "containers": {"type": "bool", "default": False},
        "containers_filters": {"type": "dict"},
        "images": {"type": "bool", "default": False},
        "images_filters": {"type": "dict"},
        "networks": {"type": "bool", "default": False},
        "networks_filters": {"type": "dict"},
        "volumes": {"type": "bool", "default": False},
        "volumes_filters": {"type": "dict"},
        "builder_cache": {"type": "bool", "default": False},
        "builder_cache_all": {"type": "bool", "default": False},
        "builder_cache_filters": {"type": "dict"},
        "builder_cache_keep_storage": {"type": "str"},  # convert to bytes
    }

    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        option_minimal_versions={
            "builder_cache": {"docker_py_version": "1.31"},
            "builder_cache_all": {"docker_py_version": "1.39"},
            "builder_cache_filters": {"docker_py_version": "1.31"},
            "builder_cache_keep_storage": {"docker_py_version": "1.39"},
        },
        # supports_check_mode=True,
    )

    builder_cache_keep_storage = None
    if client.module.params.get("builder_cache_keep_storage") is not None:
        try:
            builder_cache_keep_storage = human_to_bytes(
                client.module.params.get("builder_cache_keep_storage")
            )
        except ValueError as exc:
            client.module.fail_json(
                msg=f"Error while parsing value of builder_cache_keep_storage: {exc}"
            )

    try:
        result = {}
        changed = False

        if client.module.params["containers"]:
            filters = clean_dict_booleans_for_docker_api(
                client.module.params.get("containers_filters"), allow_sequences=True
            )
            params = {"filters": convert_filters(filters)}
            res = client.post_to_json("/containers/prune", params=params)
            result["containers"] = res.get("ContainersDeleted") or []
            result["containers_space_reclaimed"] = res["SpaceReclaimed"]
            if result["containers"] or result["containers_space_reclaimed"]:
                changed = True

        if client.module.params["images"]:
            filters = clean_dict_booleans_for_docker_api(
                client.module.params.get("images_filters"), allow_sequences=True
            )
            params = {"filters": convert_filters(filters)}
            res = client.post_to_json("/images/prune", params=params)
            result["images"] = res.get("ImagesDeleted") or []
            result["images_space_reclaimed"] = res["SpaceReclaimed"]
            if result["images"] or result["images_space_reclaimed"]:
                changed = True

        if client.module.params["networks"]:
            filters = clean_dict_booleans_for_docker_api(
                client.module.params.get("networks_filters"), allow_sequences=True
            )
            params = {"filters": convert_filters(filters)}
            res = client.post_to_json("/networks/prune", params=params)
            result["networks"] = res.get("NetworksDeleted") or []
            if result["networks"]:
                changed = True

        if client.module.params["volumes"]:
            filters = clean_dict_booleans_for_docker_api(
                client.module.params.get("volumes_filters"), allow_sequences=True
            )
            params = {"filters": convert_filters(filters)}
            res = client.post_to_json("/volumes/prune", params=params)
            result["volumes"] = res.get("VolumesDeleted") or []
            result["volumes_space_reclaimed"] = res["SpaceReclaimed"]
            if result["volumes"] or result["volumes_space_reclaimed"]:
                changed = True

        if client.module.params["builder_cache"]:
            filters = clean_dict_booleans_for_docker_api(
                client.module.params.get("builder_cache_filters"), allow_sequences=True
            )
            params = {"filters": convert_filters(filters)}
            if client.module.params.get("builder_cache_all"):
                params["all"] = "true"
            if builder_cache_keep_storage is not None:
                params["keep-storage"] = builder_cache_keep_storage
            res = client.post_to_json("/build/prune", params=params)
            result["builder_cache_space_reclaimed"] = res["SpaceReclaimed"]
            if result["builder_cache_space_reclaimed"]:
                changed = True
            if "CachesDeleted" in res:
                # API version 1.39+: return value CachesDeleted (list of str)
                result["builder_cache_caches_deleted"] = res["CachesDeleted"]
                if result["builder_cache_caches_deleted"]:
                    changed = True

        result["changed"] = changed
        client.module.exit_json(**result)
    except DockerException as e:
        client.fail(
            f"An unexpected Docker error occurred: {e}",
            exception=traceback.format_exc(),
        )
    except RequestException as e:
        client.fail(
            f"An unexpected requests error occurred when trying to talk to the Docker daemon: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# Copyright (c) 2014, Toshio Kuratomi <tkuratomi@ansible.com>
#
# Simplified BSD License (see simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import re


# Input patterns for is_input_dangerous function:
#
# 1. '"' in string and '--' in string or
# "'" in string and '--' in string
PATTERN_1 = re.compile(r'(\'|\").*--')

# 2. union \ intersect \ except + select
PATTERN_2 = re.compile(r'(UNION|INTERSECT|EXCEPT).*SELECT', re.IGNORECASE)

# 3. ';' and any KEY_WORDS
PATTERN_3 = re.compile(r';.*(SELECT|UPDATE|INSERT|DELETE|DROP|TRUNCATE|ALTER)', re.IGNORECASE)


class SQLParseError(Exception):
    pass


class UnclosedQuoteError(SQLParseError):
    pass


# maps a type of identifier to the maximum number of dot levels that are
# allowed to specify that identifier.  For example, a database column can be
# specified by up to 4 levels: database.schema.table.column
_PG_IDENTIFIER_TO_DOT_LEVEL = dict(
    database=1,
    schema=2,
    table=3,
    column=4,
    role=1,
    tablespace=1,
    sequence=3,
    publication=1,
)
_MYSQL_IDENTIFIER_TO_DOT_LEVEL = dict(database=1, table=2, column=3, role=1, vars=1)


def _find_end_quote(identifier, quote_char):
    accumulate = 0
    while True:
        try:
            quote = identifier.index(quote_char)
        except ValueError:
            raise UnclosedQuoteError
        accumulate = accumulate + quote
        try:
            next_char = identifier[quote + 1]
        except IndexError:
            return accumulate
        if next_char == quote_char:
            try:
                identifier = identifier[quote + 2:]
                accumulate = accumulate + 2
            except IndexError:
                raise UnclosedQuoteError
        else:
            return accumulate


def _identifier_parse(identifier, quote_char):
    if not identifier:
        raise SQLParseError('Identifier name unspecified or unquoted trailing dot')

    already_quoted = False
    if identifier.startswith(quote_char):
        already_quoted = True
        try:
            end_quote = _find_end_quote(identifier[1:], quote_char=quote_char) + 1
        except UnclosedQuoteError:
            already_quoted = False
        else:
            if end_quote < len(identifier) - 1:
                if identifier[end_quote + 1] == '.':
                    dot = end_quote + 1
                    first_identifier = identifier[:dot]
                    next_identifier = identifier[dot + 1:]
                    further_identifiers = _identifier_parse(next_identifier, quote_char)
                    further_identifiers.insert(0, first_identifier)
                else:
                    raise SQLParseError('User escaped identifiers must escape extra quotes')
            else:
                further_identifiers = [identifier]

    if not already_quoted:
        try:
            dot = identifier.index('.')
        except ValueError:
            identifier = identifier.replace(quote_char, quote_char * 2)
            identifier = ''.join((quote_char, identifier, quote_char))
            further_identifiers = [identifier]
        else:
            if dot == 0 or dot >= len(identifier) - 1:
                identifier = identifier.replace(quote_char, quote_char * 2)
                identifier = ''.join((quote_char, identifier, quote_char))
                further_identifiers = [identifier]
            else:
                first_identifier = identifier[:dot]
                next_identifier = identifier[dot + 1:]
                further_identifiers = _identifier_parse(next_identifier, quote_char)
                first_identifier = first_identifier.replace(quote_char, quote_char * 2)
                first_identifier = ''.join((quote_char, first_identifier, quote_char))
                further_identifiers.insert(0, first_identifier)

    return further_identifiers


def pg_quote_identifier(identifier, id_type):
    identifier_fragments = _identifier_parse(identifier, quote_char='"')
    if len(identifier_fragments) > _PG_IDENTIFIER_TO_DOT_LEVEL[id_type]:
        raise SQLParseError('PostgreSQL does not support %s with more than %i dots' % (id_type, _PG_IDENTIFIER_TO_DOT_LEVEL[id_type]))
    return '.'.join(identifier_fragments)


def mysql_quote_identifier(identifier, id_type):
    identifier_fragments = _identifier_parse(identifier, quote_char='`')
    if (len(identifier_fragments) - 1) > _MYSQL_IDENTIFIER_TO_DOT_LEVEL[id_type]:
        raise SQLParseError('MySQL does not support %s with more than %i dots' % (id_type, _MYSQL_IDENTIFIER_TO_DOT_LEVEL[id_type]))

    special_cased_fragments = []
    for fragment in identifier_fragments:
        if fragment == '`*`':
            special_cased_fragments.append('*')
        else:
            special_cased_fragments.append(fragment)

    return '.'.join(special_cased_fragments)


def is_input_dangerous(string):
    """Check if the passed string is potentially dangerous.
    Can be used to prevent SQL injections.

    Note: use this function only when you can't use
      psycopg2's cursor.execute method parametrized
      (typically with DDL queries).
    """
    if not string:
        return False

    for pattern in (PATTERN_1, PATTERN_2, PATTERN_3):
        if re.search(pattern, string):
            return True

    return False


def check_input(module, *args):
    """Wrapper for is_input_dangerous function."""
    needs_to_check = args

    dangerous_elements = []

    for elem in needs_to_check:
        if isinstance(elem, str):
            if is_input_dangerous(elem):
                dangerous_elements.append(elem)

        elif isinstance(elem, list):
            for e in elem:
                if is_input_dangerous(e):
                    dangerous_elements.append(e)

        elif elem is None or isinstance(elem, bool):
            pass

        else:
            elem = str(elem)
            if is_input_dangerous(elem):
                dangerous_elements.append(elem)

    if dangerous_elements:
        module.fail_json(msg="Passed input '%s' is "
                             "potentially dangerous" % ', '.join(dangerous_elements))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
The module file for nxos_bgp_templates
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
module: nxos_bgp_templates
short_description: BGP Templates resource module.
description:
- This module manages BGP templates on devices running Cisco NX-OS.
version_added: 4.2.0
notes:
- Tested against NX-OS 9.3.6.
- Unsupported for Cisco MDS
- This module works with connection C(network_cli) and C(httpapi).
author: Nilashish Chakraborty (@NilashishC)
options:
  running_config:
    description:
    - This option is used only with state I(parsed).
    - The value of this option should be the output received from the NX-OS device
      by executing the command B(show running-config bgp | section 'template').
    - The state I(parsed) reads the configuration from C(running_config) option and
      transforms it into Ansible structured data as per the resource module's argspec
      and the value is then returned in the I(parsed) key within the result.
    type: str
  config:
    description: A list of BGP templates.
    type: dict
    suboptions:
      as_number:
        description: Autonomous System Number of the router.
        type: str
      neighbor:
        description: Configure BGP peer templates.
        type: list
        elements: dict
        suboptions:
          name:
            description: Name of the BGP peer template.
            type: str
          address_family:
            description: Configure an address-family for peer.
            type: list
            elements: dict
            suboptions:
              afi:
                description: Address Family indicator.
                type: str
                choices: ["ipv4", "ipv6", "link-state", "l2vpn"]
                required: True
              safi:
                description: Sub Address Family indicator.
                type: str
                choices: ["unicast", "multicast", "mvpn", "evpn"]
              advertise_map:
                description: Specify route-map for conditional advertisement.
                type: dict
                suboptions:
                  route_map:
                    description: Route-map name.
                    type: str
                    required: True
                  exist_map:
                    description: Condition route-map to advertise only when prefix in condition exists.
                    type: str
                  non_exist_map:
                    description: Condition route-map to advertise only when prefix in condition does not exist.
                    type: str
              advertisement_interval:
                description: Minimum interval between sending BGP routing updates.
                type: int
              allowas_in:
                description: Accept as-path with my AS present in it.
                type: dict
                suboptions:
                  set:
                    description: Activate allowas-in property.
                    type: bool
                  max_occurences:
                    description: Number of occurrences of AS number, default is 3.
                    type: int
              as_override:
                description: Override matching AS-number while sending update.
                type: bool
              capability:
                description: Advertise capability to the peer.
                type: dict
                suboptions:
                  additional_paths:
                    description: Additional paths capability.
                    type: dict
                    suboptions:
                      receive:
                        description: Additional paths Receive capability.
                        type: str
                        choices: ["enable", "disable"]
                      send:
                        description: Additional paths Send capability.
                        type: str
                        choices: ["enable", "disable"]
              default_originate:
                description: Originate a default toward this peer.
                type: dict
                suboptions:
                  set:
                    description: Set default-originate attribute.
                    type: bool
                  route_map:
                    description: Route-map to specify criteria for originating default.
                    type: str
              disable_peer_as_check:
                description: Disable checking of peer AS-number while advertising.
                type: bool
              filter_list:
                description: Name of filter-list.
                type: dict
                suboptions:
                  inbound:
                    description: Apply policy to incoming routes.
                    type: str
                  outbound:
                    description: Apply policy to outgoing routes.
                    type: str
              inherit:
                description: Inherit a peer-policy template.
                type: dict
                suboptions:
                  peer_policy:
                    description: Peer-policy template to inherit.
                    type: str
              maximum_prefix:
                description: Maximum number of prefixes from this neighbor.
                type: dict
                suboptions:
                  max_prefix_limit:
                    description: Maximum prefix limit.
                    type: int
                  generate_warning_threshold:
                    description: Threshold percentage at which to generate a warning.
                    type: int
                  restart_interval:
                    description: Restart bgp connection after limit is exceeded.
                    type: int
                  warning_only:
                    description: Only give a warning message when limit is exceeded.
                    type: bool
              next_hop_self:
                description: Set our address as nexthop (non-reflected).
                type: dict
                suboptions:
                  set:
                    description: Set next-hop-self attribute.
                    type: bool
                  all_routes:
                    description: Set our address as nexthop for all routes.
                    type: bool
              next_hop_third_party:
                description: Compute a third-party nexthop if possible.
                type: bool
              prefix_list:
                description: Apply prefix-list.
                type: dict
                suboptions:
                  inbound:
                    description: Apply policy to incoming routes.
                    type: str
                  outbound:
                    description: Apply policy to outgoing routes.
                    type: str
              route_map:
                description: Apply route-map to neighbor.
                type: dict
                suboptions:
                  inbound:
                    description: Name of policy to apply to incoming routes.
                    type: str
                  outbound:
                    description: Name of policy to apply to outgoing routes.
                    type: str
              route_reflector_client:
                description: Configure a neighbor as Route reflector client.
                type: bool
              send_community:
                description: Send Community attribute to this neighbor.
                type: str
                choices: ["standard", "extended", "both"]
              soft_reconfiguration_inbound:
                description: Soft reconfiguration.
                type: dict
                suboptions:
                  set:
                    description: Set soft-reconfiguration inbound attribute.
                    type: bool
                  always:
                    description: Always perform inbound soft reconfiguration.
                    type: bool
              soo:
                description: Specify Site-of-origin extcommunity.
                type: str
              suppress_inactive:
                description: Advertise only active routes to peer.
                type: bool
              unsuppress_map:
                description: Route-map to selectively unsuppress suppressed routes.
                type: str
              weight:
                description: Set default weight for routes from this neighbor.
                type: int
          bfd:
            description: Bidirectional Fast Detection for the neighbor.
            type: dict
            suboptions:
              set:
                description: Set BFD for this neighbor.
                type: bool
              singlehop:
                description: Single-hop session.
                type: bool
              multihop:
                description: Multihop session.
                type: dict
                suboptions:
                  set:
                    description: Set BFD multihop.
                    type: bool
                  interval:
                    description: Configure BFD session interval parameters.
                    type: dict
                    suboptions:
                      tx_interval:
                        description: TX interval in milliseconds.
                        type: int
                        required: true
                      min_rx_interval:
                        description: Minimum RX interval.
                        type: int
                        required: true
                      multiplier:
                        description: Detect Multiplier.
                        type: int
                        required: true
          bmp_activate_server:
            description: Specify server ID for activating BMP monitoring for the peer.
            type: int
          capability:
            description: Capability.
            type: dict
            suboptions:
              suppress_4_byte_as:
                description: Suppress 4-byte AS Capability.
                type: bool
          description:
            description: Neighbor specific descripion.
            type: str
          disable_connected_check:
            description: Disable check for directly connected peer.
            type: bool
          dont_capability_negotiate:
            description: Don't negotiate capability with this neighbor.
            type: bool
          dscp:
            description: Set dscp value for tcp transport.
            type: str
          dynamic_capability:
            description: Dynamic Capability
            type: bool
          ebgp_multihop:
            description: Specify multihop TTL for remote peer.
            type: int
          graceful_shutdown:
            description: Graceful-shutdown for this neighbor.
            type: dict
            suboptions:
              activate:
                description: Send graceful-shutdown community.
                type: dict
                suboptions:
                  set:
                    description: Set activate.
                    type: bool
                  route_map:
                    description: Apply route-map to modify attributes for outbound.
                    type: str
          inherit:
            description: Inherit a template.
            type: dict
            suboptions:
              peer_session:
                description: Peer-session template to inherit.
                type: str
          local_as:
            description: Specify the local-as number for the eBGP neighbor.
            type: str
          log_neighbor_changes:
            description: Log message for neighbor up/down event.
            type: dict
            suboptions:
              set:
                description:
                - Set log-neighbor-changes.
                type: bool
              disable:
                description:
                - Disable logging of neighbor up/down event.
                type: bool
          low_memory:
            description: Behaviour in low memory situations.
            type: dict
            suboptions:
              exempt:
                description: Do not shutdown this peer when under memory pressure.
                type: bool
          password:
            description: Configure a password for neighbor.
            type: dict
            suboptions:
              encryption:
                description:
                - 0 specifies an UNENCRYPTED neighbor password.
                - 3 specifies an 3DES ENCRYPTED neighbor password will follow.
                - 7 specifies a Cisco type 7  ENCRYPTED neighbor password will follow.
                type: int
              key:
                description: Authentication password.
                type: str
          path_attribute:
            description: BGP path attribute optional filtering.
            type: list
            elements: dict
            suboptions:
              action:
                description: Action.
                type: str
                choices: ["discard", "treat-as-withdraw"]
              type:
                description: Path attribute type
                type: int
              range:
                description: Path attribute range.
                type: dict
                suboptions:
                  start:
                    description: Path attribute range start value.
                    type: int
                  end:
                    description: Path attribute range end value.
                    type: int
          remote_as:
            description: Specify Autonomous System Number of the neighbor.
            type: str
          remove_private_as:
            description: Remove private AS number from outbound updates.
            type: dict
            suboptions:
              set:
                description: Remove private AS.
                type: bool
              replace_as:
                description: Replace.
                type: bool
              all:
                description: All.
                type: bool
          shutdown:
            description: Administratively shutdown this neighbor.
            type: bool
          timers:
            description: Configure keepalive and hold timers.
            type: dict
            suboptions:
              keepalive:
                description: Keepalive interval (seconds).
                type: int
              holdtime:
                description: Holdtime (seconds).
                type: int
          transport:
            description: BGP transport connection.
            type: dict
            suboptions:
              connection_mode:
                description: Specify type of connection.
                type: dict
                suboptions:
                  passive:
                    description: Allow passive connection setup only.
                    type: bool
          ttl_security:
            description: Enable TTL Security Mechanism.
            type: dict
            suboptions:
              hops:
                description: Specify hop count for remote peer.
                type: int
          update_source:
            description: Specify source of BGP session and updates.
            type: str
  state:
    description:
    - The state the configuration should be left in.
    - Refer to examples for more details.
    type: str
    choices:
    - merged
    - replaced
    - overridden
    - deleted
    - parsed
    - gathered
    - rendered
    default: merged
"""

EXAMPLES = """
# Using merged

# Before state:
# --------------
#
# nxos9k# show running-config | section "^router bgp"
# nxos9k#

- name: Merge the provided configuration with the existing running configuration
  cisco.nxos.nxos_bgp_templates:
    config:
      as_number: 65536
      neighbor:
        - name: neighbor_tmplt_1
          address_family:
            - afi: ipv4
              safi: unicast
              advertise_map:
                route_map: rmap1
                non_exist_map: nemap1
              advertisement_interval: 60
              disable_peer_as_check: true
          bmp_activate_server: 2
          capability:
            suppress_4_byte_as: true
          description: Test_BGP_PEER_TEMPLATE_1
          local_as: 65536
          remote_as: 65001

        - name: neighbor_tmplt_2
          description: Test_BGP_PEER_TEMPLATE_2
          address_family:
            - afi: ipv4
              safi: multicast
              advertise_map:
                route_map: rmap1
                exist_map: emap1
              as_override: true
              filter_list:
                inbound: flist1
                outbound: flist2
          inherit:
            peer_session: psession1
          timers:
            holdtime: 100
            keepalive: 45
# Task Output:
# ------------
# before: {}
#
# commands:
#   - router bgp 65536
#   - template peer neighbor_tmplt_1
#   - bmp-activate-server 2
#   - capability suppress 4-byte-as
#   - description Test_BGP_PEER_TEMPLATE_1
#   - local-as 65536
#   - remote-as 65001
#   - address-family ipv4 unicast
#   - advertise-map rmap1 non-exist-map nemap1
#   - advertisement-interval 60
#   - disable-peer-as-check
#   - template peer neighbor_tmplt_2
#   - description Test_BGP_PEER_TEMPLATE_2
#   - inherit peer-session psession1
#   - timers 45 100
#   - address-family ipv4 multicast
#   - advertise-map rmap1 exist-map emap1
#   - as-override
#   - filter-list flist1 in
#   - filter-list flist2 out
#
# after:
#   as_number: "65536"
#   neighbor:
#     - name: neighbor_tmplt_1
#       address_family:
#         - afi: ipv4
#           safi: unicast
#           advertise_map:
#             non_exist_map: nemap1
#             route_map: rmap1
#           advertisement_interval: 60
#           disable_peer_as_check: true
#       bmp_activate_server: 2
#       capability:
#         suppress_4_byte_as: true
#       description: Test_BGP_PEER_TEMPLATE_1
#       local_as: "65536"
#       remote_as: "65001"
#
#     - name: neighbor_tmplt_2
#       description: Test_BGP_PEER_TEMPLATE_2
#       address_family:
#         - afi: ipv4
#           safi: multicast
#           advertise_map:
#             exist_map: emap1
#             route_map: rmap1
#           as_override: true
#           filter_list:
#             inbound: flist1
#             outbound: flist2
#       inherit:
#         peer_session: psession1
#       timers:
#         holdtime: 100
#         keepalive: 45

# After state:
# --------------
#
# nxos9k# show running-config | section "^router bgp"
# router bgp 65536
#   template peer neighbor_tmplt_1
#     capability suppress 4-byte-as
#     bmp-activate-server 2
#     description Test_BGP_PEER_TEMPLATE_1
#     local-as 65536
#     remote-as 65001
#     address-family ipv4 unicast
#      advertise-map rmap1 non-exist-map nemap1
#      advertisement-interval 60
#      disable-peer-as-check
#   template peer neighbor_tmplt_2
#     description Test_BGP_PEER_TEMPLATE_2
#     inherit peer-session psession1
#     timers 45 100
#     address-family ipv4 multicast
#      advertise-map rmap1 exist-map emap1
#      as-override
#      filter-list flist1 in
#      filter-list flist2 out

# Using replaced

# Before state:
# -------------
#
# nxos9k# show running-config | section "^router bgp"
# router bgp 65536
#   template peer neighbor_tmplt_1
#     capability suppress 4-byte-as
#     description Test_BGP_PEER_TEMPLATE_1
#     bmp-activate-server 2
#     local-as 65536
#     remote-as 65001
#     address-family ipv4 unicast
#      advertise-map rmap1 non-exist-map nemap1
#      advertisement-interval 60
#      disable-peer-as-check
#   template peer neighbor_tmplt_2
#     description Test_BGP_PEER_TEMPLATE_2
#     inherit peer-session psession1
#     timers 45 100
#     address-family ipv4 multicast
#      advertise-map rmap1 exist-map emap1
#      as-override
#      filter-list flist1 in
#      filter-list flist2 out

- name: Replace BGP templates configuration with provided configuration
  cisco.nxos.nxos_bgp_templates:
    config:
      as_number: 65536
      neighbor:
        - name: neighbor_tmplt_1
          address_family:
            - afi: ipv4
              safi: unicast
              advertise_map:
                route_map: rmap1
                non_exist_map: nemap1
              advertisement_interval: 60
              disable_peer_as_check: true
          inherit:
            peer_session: psession1
          description: Test_BGP_PEER_TEMPLATE_1
          local_as: 65537
    state: replaced

# Task output:
# ------------
#
# before:
#   as_number: "65536"
#   neighbor:
#     - name: neighbor_tmplt_1
#       address_family:
#         - afi: ipv4
#           safi: unicast
#           advertise_map:
#             non_exist_map: nemap1
#             route_map: rmap1
#           advertisement_interval: 60
#           disable_peer_as_check: true
#       bmp_activate_server: 2
#       capability:
#         suppress_4_byte_as: true
#       description: Test_BGP_PEER_TEMPLATE_1
#       local_as: "65536"
#       remote_as: "65001"
#
#     - name: neighbor_tmplt_2
#       description: Test_BGP_PEER_TEMPLATE_2
#       address_family:
#         - afi: ipv4
#           safi: multicast
#           advertise_map:
#             exist_map: emap1
#             route_map: rmap1
#           as_override: true
#           filter_list:
#             inbound: flist1
#             outbound: flist2
#       inherit:
#         peer_session: psession1
#       timers:
#         holdtime: 100
#         keepalive: 45
#
# commands:
#   - router bgp 65536
#   - template peer neighbor_tmplt_1
#   - no bmp-activate-server 2
#   - no capability suppress 4-byte-as
#   - inherit peer-session psession1
#   - local-as 65537
#   - no remote-as 65001
#
# after:
#   as_number: "65536"
#   neighbor:
#     - name: neighbor_tmplt_1
#       address_family:
#         - afi: ipv4
#           safi: unicast
#           advertise_map:
#             non_exist_map: nemap1
#             route_map: rmap1
#           advertisement_interval: 60
#           disable_peer_as_check: true
#       description: Test_BGP_PEER_TEMPLATE_1
#       inherit:
#         peer_session: psession1
#       local_as: "65537"
#
#     - name: neighbor_tmplt_2
#       description: Test_BGP_PEER_TEMPLATE_2
#       address_family:
#         - afi: ipv4
#           safi: multicast
#           advertise_map:
#             exist_map: emap1
#             route_map: rmap1
#           as_override: true
#           filter_list:
#             inbound: flist1
#             outbound: flist2
#       inherit:
#         peer_session: psession1
#       timers:
#         holdtime: 100
#         keepalive: 45

# After state:
# ------------
#
# nxos9k# show running-config | section "^router bgp"
# router bgp 65536
#   template peer neighbor_tmplt_1
#     inherit peer-session psession1
#     description Test_BGP_PEER_TEMPLATE_1
#     local-as 65537
#     address-family ipv4 unicast
#      advertise-map rmap1 non-exist-map nemap1
#      advertisement-interval 60
#      disable-peer-as-check
#   template peer neighbor_tmplt_2
#     description Test_BGP_PEER_TEMPLATE_2
#     inherit peer-session psession1
#     bmp-activate-server 2
#     timers 45 100
#     address-family ipv4 multicast
#      advertise-map rmap1 exist-map emap1
#      as-override
#      filter-list flist1 in
#      filter-list flist2 out

# Using overridden
#
# Before state:
# -------------
#
# nxos9k# show running-config | section "^router bgp"
# router bgp 65536
#   template peer neighbor_tmplt_1
#     capability suppress 4-byte-as
#     description Test_BGP_PEER_TEMPLATE_1
#     bmp-activate-server 2
#     local-as 65536
#     remote-as 65001
#     address-family ipv4 unicast
#      advertise-map rmap1 non-exist-map nemap1
#      advertisement-interval 60
#      disable-peer-as-check
#   template peer neighbor_tmplt_2
#     description Test_BGP_PEER_TEMPLATE_2
#     inherit peer-session psession1
#     timers 45 100
#     address-family ipv4 multicast
#      advertise-map rmap1 exist-map emap1
#      as-override
#      filter-list flist1 in
#      filter-list flist2 out

- name: Override BGP templates configuration with provided configuration
  cisco.nxos.nxos_bgp_templates:
    config:
      as_number: 65536
      neighbor:
        - name: neighbor_tmplt_1
          address_family:
            - afi: ipv4
              safi: unicast
              advertise_map:
                route_map: rmap1
                non_exist_map: nemap1
              advertisement_interval: 60
              disable_peer_as_check: true
          inherit:
            peer_session: psession1
          description: Test_BGP_PEER_TEMPLATE_1
          local_as: 65537
    state: overridden

# Task output:
# ------------
#
# before:
#   as_number: "65536"
#   neighbor:
#     - name: neighbor_tmplt_1
#       address_family:
#         - afi: ipv4
#           safi: unicast
#           advertise_map:
#             non_exist_map: nemap1
#             route_map: rmap1
#           advertisement_interval: 60
#           disable_peer_as_check: true
#       bmp_activate_server: 2
#       capability:
#         suppress_4_byte_as: true
#       description: Test_BGP_PEER_TEMPLATE_1
#       local_as: "65536"
#       remote_as: "65001"
#
#     - name: neighbor_tmplt_2
#       description: Test_BGP_PEER_TEMPLATE_2
#       address_family:
#         - afi: ipv4
#           safi: multicast
#           advertise_map:
#             exist_map: emap1
#             route_map: rmap1
#           as_override: true
#           filter_list:
#             inbound: flist1
#             outbound: flist2
#       inherit:
#         peer_session: psession1
#       timers:
#         holdtime: 100
#         keepalive: 45
#
# commands:
#   - router bgp 65536
#   - template peer neighbor_tmplt_1
#   - no bmp-activate-server 2
#   - no capability suppress 4-byte-as
#   - inherit peer-session psession1
#   - local-as 65537
#   - no remote-as 65001
#   - no template peer neighbor_tmplt_2
#
# after:
#   as_number: "65536"
#   neighbor:
#     - name: neighbor_tmplt_1
#       address_family:
#         - afi: ipv4
#           safi: unicast
#           advertise_map:
#             non_exist_map: nemap1
#             route_map: rmap1
#           advertisement_interval: 60
#           disable_peer_as_check: true
#       description: Test_BGP_PEER_TEMPLATE_1
#       inherit:
#         peer_session: psession1
#       local_as: "65537"

# After state:
# ------------
#
# nxos9k# show running-config | section "^router bgp"
# router bgp 65536
#   template peer neighbor_tmplt_1
#     inherit peer-session psession1
#     description Test_BGP_PEER_TEMPLATE_1
#     local-as 65537
#     address-family ipv4 unicast
#      advertise-map rmap1 non-exist-map nemap1
#      advertisement-interval 60
#      disable-peer-as-check

# Using deleted

# Before state:
# --------------
#
# nxos9k# show running-config | section "^router bgp"
# router bgp 65536
#   template peer neighbor_tmplt_1
#     capability suppress 4-byte-as
#     description Test_BGP_PEER_TEMPLATE_1
#     bmp-activate-server 2
#     local-as 65536
#     remote-as 65001
#     address-family ipv4 unicast
#      advertise-map rmap1 non-exist-map nemap1
#      advertisement-interval 60
#      disable-peer-as-check
#   template peer neighbor_tmplt_2
#     description Test_BGP_PEER_TEMPLATE_2
#     inherit peer-session psession1
#     timers 45 100
#     address-family ipv4 multicast
#      advertise-map rmap1 exist-map emap1
#      as-override
#      filter-list flist1 in
#      filter-list flist2 out

- name: Delete BGP configs handled by this module
  cisco.nxos.nxos_bgp_templates:
    state: deleted

# Task output:
# ------------
#
# before:
#   as_number: "65536"
#   neighbor:
#     - name: neighbor_tmplt_1
#       address_family:
#         - afi: ipv4
#           safi: unicast
#           advertise_map:
#             non_exist_map: nemap1
#             route_map: rmap1
#           advertisement_interval: 60
#           disable_peer_as_check: true
#       bmp_activate_server: 2
#       capability:
#         suppress_4_byte_as: true
#       description: Test_BGP_PEER_TEMPLATE_1
#       local_as: "65536"
#       remote_as: "65001"
#
#     - name: neighbor_tmplt_2
#       description: Test_BGP_PEER_TEMPLATE_2
#       address_family:
#         - afi: ipv4
#           safi: multicast
#           advertise_map:
#             exist_map: emap1
#             route_map: rmap1
#           as_override: true
#           filter_list:
#             inbound: flist1
#             outbound: flist2
#       inherit:
#         peer_session: psession1
#       timers:
#         holdtime: 100
#         keepalive: 45
#
# commands:
#   - router bgp 65536
#   - no template peer neighbor_tmplt_1
#   - no template peer neighbor_tmplt_2
#
# after: {}

# After state:
# -------------
# nxos9k# show running-config | section "^router bgp"
# nxos9k#
"""

RETURN = """
before:
  description: The configuration prior to the module execution.
  returned: when I(state) is C(merged), C(replaced), C(overridden), C(deleted) or C(purged)
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
after:
  description: The resulting configuration after module execution.
  returned: when changed
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
commands:
  description: The set of commands pushed to the remote device.
  returned: when I(state) is C(merged), C(replaced), C(overridden), C(deleted) or C(purged)
  type: list
  sample:
    - router bgp 65536
    - template peer neighbor_tmplt_1
    - no bmp-activate-server 2
    - no capability suppress 4-byte-as
    - inherit peer-session psession1
    - local-as 65537
    - no remote-as 65001
    - no template peer neighbor_tmplt_2
rendered:
  description: The provided configuration in the task rendered in device-native format (offline).
  returned: when I(state) is C(rendered)
  type: list
  sample:
    - router bgp 65536
    - template peer neighbor_tmplt_1
    - bmp-activate-server 2
    - no capability suppress 4-byte-as
    - no template peer neighbor_tmplt_2
gathered:
  description: Facts about the network resource gathered from the remote device as structured data.
  returned: when I(state) is C(gathered)
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
parsed:
  description: The device native config provided in I(running_config) option parsed into structured data as per module argspec.
  returned: when I(state) is C(parsed)
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.argspec.bgp_templates.bgp_templates import (
    Bgp_templatesArgs,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.config.bgp_templates.bgp_templates import (
    Bgp_templates,
)


def main():
    """
    Main entry point for module execution

    :returns: the result form module invocation
    """
    module = AnsibleModule(
        argument_spec=Bgp_templatesArgs.argument_spec,
        mutually_exclusive=[["config", "running_config"]],
        required_if=[
            ["state", "merged", ["config"]],
            ["state", "replaced", ["config"]],
            ["state", "overridden", ["config"]],
            ["state", "rendered", ["config"]],
            ["state", "parsed", ["running_config"]],
        ],
        supports_check_mode=True,
    )

    result = Bgp_templates(module).execute_module()
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#
# Copyright (c) 2023 Felix Fontein <felix@fontein.de>
# Simplified BSD License (see LICENSES/BSD-2-Clause.txt or https://opensource.org/licenses/BSD-2-Clause)
# SPDX-License-Identifier: BSD-2-Clause

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import datetime as _datetime


def ensure_timezone_info(value: _datetime.datetime) -> _datetime.datetime:
    if value.tzinfo is not None:
        return value
    return value.astimezone(_datetime.timezone.utc)


def fromtimestamp(value: int | float) -> _datetime.datetime:
    return _datetime.datetime.fromtimestamp(value, tz=_datetime.timezone.utc)


def now() -> _datetime.datetime:
    return _datetime.datetime.now(tz=_datetime.timezone.utc)



#
# -*- coding: utf-8 -*-
# Copyright 2025 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The nxos_interfaces config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.facts import (
    Facts,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.rm_templates.interfaces import (
    InterfacesTemplate,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.utils.utils import (
    normalize_interface,
)


class Interfaces(ResourceModule):
    """
    The nxos_interfaces config class
    """

    def __init__(self, module):
        super(Interfaces, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="interfaces",
            tmplt=InterfacesTemplate(),
        )
        self.parsers = [
            "description",
            "speed",
            "mtu",
            "duplex",
            "ip_forward",
            "fabric_forwarding_anycast_gateway",
            "mac_address",
            "logging.link_status",
            "logging.trunk_status",
            "snmp.trap.link_status",
            "service_policy.input",
            "service_policy.output",
            "service_policy.type_options.qos.input",
            "service_policy.type_options.qos.output",
            "service_policy.type_options.queuing.input",
            "service_policy.type_options.queuing.output",
        ]
        if self.state not in ["parsed", "rendered"]:
            self.defaults = self.get_interface_defaults()
        else:
            # For parsed/rendered state, we assume defaults
            self.defaults = {
                "default_mode": "layer3",
                "L2_enabled": True,
            }

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def get_switchport_defaults(self):
        """Wrapper method for `_connection.get()`
        This method exists solely to allow the unit test framework to mock device connection calls.
        """
        return self._connection.get(
            "show running-config all | incl 'system default switchport'",
        )

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        wantd = {entry["name"]: entry for entry in self.want}
        haved = {entry["name"]: entry for entry in self.have}

        for each in wantd, haved:
            self.normalize_interface_names(each)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state == "deleted":
            haved = {k: v for k, v in haved.items() if k in wantd or not wantd}
            wantd = {}

        # remove superfluous config for overridden and deleted
        if self.state in ["overridden", "deleted"]:
            for k, have in haved.items():
                if k not in wantd:
                    self._compare(want={}, have=have)

        if self.state == "purged":
            self.purge(wantd, haved)
        else:
            for k, want in wantd.items():
                self._compare(want=want, have=haved.pop(k, {}))

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Interfaces network resource.
        """
        begin = len(self.commands)
        self.compare(parsers=self.parsers, want=want, have=have)

        # Handle the 'enabled' state separately
        want_enabled = want.get("enabled")
        have_enabled = have.get("enabled")
        if want_enabled is not None:
            if want_enabled != have_enabled:
                if want_enabled is True:
                    self.addcmd(want, "enabled", True)
                else:
                    self.addcmd(want, "enabled", False)
        elif not want and self.state == "overridden":
            if have_enabled is not None:
                self.addcmd(have, "enabled", False)
        elif not want and self.state == "deleted":
            if have_enabled:
                self.addcmd(have, "enabled", False)

        # Handle the 'mode' state separately
        want_mode = want.get("mode")
        have_mode = have.get("mode", self.defaults.get("default_mode"))
        if want_mode is not None:
            if want_mode != have_mode:
                if want_mode == "layer3":
                    self.addcmd(want, "mode", True)
                else:
                    if want:
                        self.addcmd(want, "mode", False)
                    elif have.get("mode"):  # can oly have layer2 as switchport no show cli
                        # handles deleted as want be blank and only
                        self.addcmd(have, "mode", False)
        elif not want and self.state == "deleted":
            if have.get("mode") and have_mode != self.defaults.get("default_mode"):
                no_cmd = True if self.defaults.get("default_mode") == "layer3" else False
                self.addcmd(have, "mode", no_cmd)

        if len(self.commands) != begin:
            self.commands.insert(begin, self._tmplt.render(want or have, "name", False))

    def purge(self, wantd, haved):
        """Handle operation for purged state"""
        if wantd:
            for k, want in wantd.items():
                have = haved.pop(k, {})
                if have:
                    self.commands.append(self._tmplt.render(have, "name", True))

    def normalize_interface_names(self, param):
        if param:
            for _k, val in param.items():
                val["name"] = normalize_interface(val["name"])
        return param

    def get_interface_defaults(self):
        """Collect user-defined-default states for 'system default switchport'
        configurations. These configurations determine default L2/L3 modes
        and enabled/shutdown states. The default values for user-defined-default
        configurations may be different for legacy platforms.
        Notes:
        - L3 enabled default state is False on N9K,N7K,N3K but True for N5K,N6K
        - Changing L2-L3 modes may change the default enabled value.
        - '(no) system default switchport shutdown' only applies to L2 interfaces.
        Run through the gathered interfaces and tag their default enabled state.
        """
        interface_defs = {}
        switchport_data = self.get_switchport_defaults()

        # Layer 2/3 mode defaults
        pat = "(no )*system default switchport$"
        default_mode = re.search(pat, switchport_data, re.MULTILINE)
        if default_mode:
            interface_defs["default_mode"] = (
                "layer3" if "no " in default_mode.groups() else "layer2"
            )

        # Interface enabled state defaults
        pat = "(no )*system default switchport shutdown$"
        default_enabled = re.search(pat, switchport_data, re.MULTILINE)
        if default_enabled:
            interface_defs["L2_enabled"] = True if "no " in default_enabled.groups() else False

        return interface_defs



# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Static_routes parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


class Static_routesTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Static_routesTemplate, self).__init__(
            lines=lines,
            tmplt=self,
            module=module,
        )

    # fmt: off
    PARSERS = [
        #  this parser shared the vrf context names with routes
        {
            'name': 'vrf',
            'getval': re.compile(
                r'''
                ^vrf\scontext\s(?P<namevrf>\S+)$
                ''', re.VERBOSE,
            ),
            'setval': 'vrf context {{ namevrf }}',
            'result': {},
            'shared': True,
        },
        {
            "name": "ipv4",
            "getval": re.compile(
                r"""
                (^|\s+)ip\sroute
                (\s(?P<dest>\S+))
                (\s(?P<interface>(Ethernet|loopback|mgmt|Null|port-channel)\S+))?
                (\s(?P<forward_router_address>\S+))?
                (\svrf\s(?P<dest_vrf>\S+))?
                (\strack\s(?P<track>\d+))?
                (\sname\s(?P<route_name>\S+))?
                (\stag\s(?P<tag>\d+))?
                (\s(?P<admin_distance>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "ip route"
            "{{ (' ' + ipv4.dest) if ipv4.dest is defined else '' }}"
            "{{ (' ' + ipv4.interface) if ipv4.interface is defined else '' }}"
            "{{ (' ' + ipv4.forward_router_address) if ipv4.forward_router_address is defined else '' }}"
            "{{ (' vrf ' + ipv4.dest_vrf) if ipv4.dest_vrf is defined else '' }}"
            "{{ (' track ' + ipv4.track|string) if ipv4.track is defined else '' }}"
            "{{ (' name ' + ipv4.route_name) if ipv4.route_name is defined else '' }}"
            "{{ (' tag ' + ipv4.tag|string) if ipv4.tag is defined else '' }}"
            "{{ (' ' + ipv4.admin_distance|string) if ipv4.admin_distance is defined else '' }}",
            "result": {
                "{{ dest }}_{{ namevrf|d() }}_ipv4": [
                    {
                        "_vrf": "{{ namevrf }}",
                        "_afi": "ipv4",
                        "_dest": "{{ dest }}",
                        "interface": "{{ interface }}",
                        "forward_router_address": "{{ forward_router_address }}",
                        "admin_distance": "{{ admin_distance }}",
                        "dest_vrf": "{{ dest_vrf }}",
                        "tag": "{{ tag }}",
                        "route_name": "{{ route_name }}",
                        "track": "{{ track }}",
                    },
                ],
            },
        },
        {
            "name": "ipv6",
            "getval": re.compile(
                r"""
                (^|\s+)ipv6\sroute
                (\s(?P<dest>\S+))
                (\s(?P<interface>(Ethernet|loopback|mgmt|Null|port-channel)\S+))?
                (\s(?P<forward_router_address>\S+))?
                (\svrf\s(?P<dest_vrf>\S+))?
                (\strack\s(?P<track>\d+))?
                (\sname\s(?P<route_name>\S+))?
                (\stag\s(?P<tag>\d+))?
                (\s(?P<admin_distance>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "ipv6 route"
            "{{ (' ' + ipv6.dest) if ipv6.dest is defined else '' }}"
            "{{ (' ' + ipv6.interface) if ipv6.interface is defined else '' }}"
            "{{ (' ' + ipv6.forward_router_address) if ipv6.forward_router_address is defined else '' }}"
            "{{ (' vrf ' + ipv6.dest_vrf) if ipv6.dest_vrf is defined else '' }}"
            "{{ (' track ' + ipv6.track|string) if ipv6.track is defined else '' }}"
            "{{ (' name ' + ipv6.route_name) if ipv6.route_name is defined else '' }}"
            "{{ (' tag ' + ipv6.tag|string) if ipv6.tag is defined else '' }}"
            "{{ (' ' + ipv6.admin_distance|string) if ipv6.admin_distance is defined else '' }}",
            "result": {
                "{{ dest }}_{{ namevrf|d() }}_ipv6": [
                    {
                        "_vrf": "{{ namevrf }}",
                        "_afi": "ipv6",
                        "_dest": "{{ dest }}",
                        "interface": "{{ interface }}",
                        "forward_router_address": "{{ forward_router_address }}",
                        "admin_distance": "{{ admin_distance }}",
                        "dest_vrf": "{{ dest_vrf }}",
                        "tag": "{{ tag }}",
                        "route_name": "{{ route_name }}",
                        "track": "{{ track }}",
                    },
                ],
            },
        },
    ]
    # fmt: on



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_postgresqlflexibleadministrator
version_added: "3.6.0"
short_description: Manage PostgreSQL Flexible Administrator instance
description:
    - Add or Delete instance of PostgreSQL Flexible Administrator.
    - Require the postflexible server Enable Microsoft Entra authenticate.

options:
    resource_group:
        description:
            - The name of the resource group that contains the resource. You can obtain this value from the Azure Resource Manager API or the portal.
        required: True
        type: str
    server_name:
        description:
            - The name of the post gresql flexible server.
        required: True
        type: str
    object_id:
        description:
            - The Object ID of Azure Directory Administrator.
        required: True
        type: str
    principal_name:
        description:
            - Active Directory administrator principal name.
        type: str
    principal_type:
        description:
            - The principal type used to represent the type of Active Directory Administrator.
        type: str
        choices:
            - Unknown
            - User
            - Group
            - ServicePrincipal
    tenant_id:
        description:
            - The tenant ID of Active Directory Administrator.
        type: str
    state:
        description:
            - Assert the state of the PostgreSQL Flexible administrator.
            - Use C(present) to create or update a administrator and C(absent) to delete it.
        default: present
        type: str
        choices:
            - absent
            - present

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - magodo (@magodo)
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)

'''

EXAMPLES = '''
- name: Create (or update) PostgreSQL Flexible Administrator
  azure_rm_postgresqlflexibleadministrator:
    resource_group: myResourceGroup
    server_name: testserver
    object_id: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    principal_type: User
    principal_name: fred-sun
    tenant_id: yyyyyyy-yyyy-yyyy-yyyy-yyyyyyyyyyyy

- name: Delete PostgreSQL Flexible Administrator
  azure_rm_postgresqlflexibleadministrator:
    resource_group: myResourceGroup
    server_name: testserver
    object_id: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    state: absent
'''

RETURN = '''
administrator:
    description:
        - A list of dictionaries containing facts for PostgreSQL Flexible Administrator.
    returned: always
    type: complex
    contains:
        id:
            description:
                - Resource ID of the postgresql flexible admistrator.
            returned: always
            type: str
            sample: "/subscriptions/xxx-xxx/resourceGroups/testRG/providers/Microsoft.DBforPostgreSQL/flexibleServers/postgresql03/administrators/xxx-xxx"
        principal_name:
            description:
                - Active Directory administrator principal name.
            returned: always
            type: str
            sample: fred-sun
        principal_type:
            description:
                - The principal type used to represent the type of Active Directory Administrator.
            returned: always
            type: str
            sample: User
        object_id:
            description:
                - The Object ID of Azure Directory Administrator.
            returned: always
            type: str
            sample: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
        type:
            description:
                - The type of the resource.
            returned: always
            type: str
            sample: Microsoft.DBforPostgreSQL/flexibleServers/administrators
        tenant_id:
            description:
                - The tenant ID of Active Directory Administrator.
            returned: always
            type: str
            sample: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxxx
        resource_group:
            description:
                - The resource group name.
            returned: always
            type: str
            sample: testRG
        server_name:
            description:
                - The type of the resource.
            returned: always
            type: str
            sample: postgresql03
'''


try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
    from azure.core.exceptions import ResourceNotFoundError
    from azure.core.polling import LROPoller
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMPostgreSqlFlexibleAdministrator(AzureRMModuleBase):
    """Configuration class for an Azure RM PostgreSQL Flexible Administrator resource"""

    def __init__(self):
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str',
                required=True
            ),
            server_name=dict(
                type='str',
                required=True
            ),
            object_id=dict(
                type='str',
                required=True
            ),
            principal_type=dict(
                type='str',
                choices=['Unknown', 'User', 'Group', 'ServicePrincipal']
            ),
            principal_name=dict(
                type='str'
            ),
            tenant_id=dict(
                type='str'
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            )
        )

        self.resource_group = None
        self.server_name = None
        self.object_id = None
        self.parameters = dict()

        self.results = dict(changed=False)
        self.state = None

        super(AzureRMPostgreSqlFlexibleAdministrator, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                                     supports_check_mode=True,
                                                                     supports_tags=False)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in list(self.module_arg_spec.keys()):
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
            else:
                self.parameters[key] = kwargs[key]

        old_response = None
        response = None
        changed = False

        old_response = self.get_postgresqlflexibleadministrator()

        if not old_response:
            self.log("PostgreSQL Flexible Administrator instance doesn't exist")
            if self.state == 'absent':
                self.log("The PostgreSQL Flexible Administrator with object {0} no exist".format(self.object_id))
            else:
                changed = True
                if not self.check_mode:
                    response = self.add_postgresqlflexibleadministrator(self.parameters)
        else:
            self.log("PostgreSQL Flexible Administrator instance with object {0} already exists".format(self.object_id))
            if self.state == 'absent':
                changed = True
                if not self.check_mode:
                    response = self.delete_postgresqlflexibleadministrator()
            else:
                if self.check_mode:
                    changed = True
                    self.fail("PostgreSQL Flexible Administrator instance with object {0} already exists".format(self.object_id))
                response = old_response

        self.results['administrator'] = response
        self.results['changed'] = changed
        return self.results

    def add_postgresqlflexibleadministrator(self, body):
        '''
        Add PostgreSQL Flexible Administrator with the specified configuration.

        :return: deserialized PostgreSQL Flexible Administrator instance state dictionary
        '''
        self.log("Adding the PostgreSQL Flexible Administrator instance {0}".format(self.object_id))

        try:
            response = self.postgresql_flexible_client.administrators.begin_create(resource_group_name=self.resource_group,
                                                                                   server_name=self.server_name,
                                                                                   object_id=self.object_id,
                                                                                   parameters=body)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)

        except Exception as exc:
            self.log('Error attempting to add the PostgreSQL Flexible Administrator instance.')
            self.fail("Error add the PostgreSQL Flexible Administrator instance: {0}".format(str(exc)))
        return self.format_item(response)

    def delete_postgresqlflexibleadministrator(self):
        '''
        Deletes specified PostgreSQL Flexible Administrator instance in the specified server name and resource group.

        :return: True
        '''
        self.log("Deleting the PostgreSQL Flexible Administrator instance {0}".format(self.object_id))
        try:
            self.postgresql_flexible_client.administrators.begin_delete(resource_group_name=self.resource_group,
                                                                        server_name=self.server_name,
                                                                        object_id=self.object_id)
        except Exception as ec:
            self.log('Error attempting to delete the PostgreSQL Flexible Administrator instance.')
            self.fail("Error deleting the PostgreSQL Flexible Administrator instance: {0}".format(str(ec)))

    def get_postgresqlflexibleadministrator(self):
        '''
        Gets the properties of the specified PostgreSQL Flexible Administrator.

        :return: deserialized PostgreSQL Flexible Administrator instance state dictionary
        '''
        self.log("Checking if the PostgreSQL Flexible Administrator instance {0} is present".format(self.object_id))
        found = False
        try:
            response = self.postgresql_flexible_client.administrators.get(resource_group_name=self.resource_group,
                                                                          server_name=self.server_name,
                                                                          object_id=self.object_id)
            found = True
            self.log("Response : {0}".format(response))
            self.log("PostgreSQL Flexible Administrator instance : {0} found".format(response.name))
        except ResourceNotFoundError as e:
            self.log('Did not find the PostgreSQL Flexible Administrator instance. Exception as {0}'.format(e))
        if found is True:
            return self.format_item(response)

        return None

    def format_item(self, item):
        return dict(
            resource_group=self.resource_group,
            server_name=self.server_name,
            object_id=item.object_id,
            id=item.id,
            principal_name=item.principal_name,
            principal_type=item.principal_type,
            tenant_id=item.tenant_id,
            type=item.type
        )


def main():
    """Main execution"""
    AzureRMPostgreSqlFlexibleAdministrator()


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2014, Michael Warkentin <mwarkentin@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: bower
short_description: Manage bower packages with C(bower)
description:
  - Manage bower packages with C(bower).
author: "Michael Warkentin (@mwarkentin)"
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  name:
    type: str
    description:
      - The name of a bower package to install.
  offline:
    description:
      - Install packages from local cache, if the packages were installed before.
    type: bool
    default: false
  production:
    description:
      - Install with C(--production) flag.
    type: bool
    default: false
  path:
    type: path
    description:
      - The base path where to install the bower packages.
    required: true
  relative_execpath:
    type: path
    description:
      - Relative path to bower executable from install path.
  state:
    type: str
    description:
      - The state of the bower package.
    default: present
    choices: ["present", "absent", "latest"]
  version:
    type: str
    description:
      - The version to be installed.
"""

EXAMPLES = r"""
- name: Install "bootstrap" bower package.
  community.general.bower:
    name: bootstrap

- name: Install "bootstrap" bower package on version 3.1.1.
  community.general.bower:
    name: bootstrap
    version: '3.1.1'

- name: Remove the "bootstrap" bower package.
  community.general.bower:
    name: bootstrap
    state: absent

- name: Install packages based on bower.json.
  community.general.bower:
    path: /app/location

- name: Update packages based on bower.json to their latest version.
  community.general.bower:
    path: /app/location
    state: latest

# install bower locally and run from there
- npm:
    path: /app/location
    name: bower
    global: false
- community.general.bower:
    path: /app/location
    relative_execpath: node_modules/.bin
"""

import json
import os

from ansible.module_utils.basic import AnsibleModule


class Bower:
    def __init__(self, module, **kwargs):
        self.module = module
        self.name = kwargs["name"]
        self.offline = kwargs["offline"]
        self.production = kwargs["production"]
        self.path = kwargs["path"]
        self.relative_execpath = kwargs["relative_execpath"]
        self.version = kwargs["version"]

        if kwargs["version"]:
            self.name_version = f"{self.name}#{self.version}"
        else:
            self.name_version = self.name

    def _exec(self, args, run_in_check_mode=False, check_rc=True):
        if not self.module.check_mode or (self.module.check_mode and run_in_check_mode):
            cmd = []

            if self.relative_execpath:
                cmd.append(os.path.join(self.path, self.relative_execpath, "bower"))
                if not os.path.isfile(cmd[-1]):
                    self.module.fail_json(msg=f"bower not found at relative path {self.relative_execpath}")
            else:
                cmd.append("bower")

            cmd.extend(args)
            cmd.extend(["--config.interactive=false", "--allow-root"])

            if self.name:
                cmd.append(self.name_version)

            if self.offline:
                cmd.append("--offline")

            if self.production:
                cmd.append("--production")

            # If path is specified, cd into that path and run the command.
            cwd = None
            if self.path:
                if not os.path.exists(self.path):
                    os.makedirs(self.path)
                if not os.path.isdir(self.path):
                    self.module.fail_json(msg=f"path {self.path} is not a directory")
                cwd = self.path

            rc, out, err = self.module.run_command(cmd, check_rc=check_rc, cwd=cwd)
            return out
        return ""

    def list(self):
        cmd = ["list", "--json"]

        installed = list()
        missing = list()
        outdated = list()
        data = json.loads(self._exec(cmd, True, False))
        if "dependencies" in data:
            for dep in data["dependencies"]:
                dep_data = data["dependencies"][dep]
                if dep_data.get("missing", False):
                    missing.append(dep)
                elif (
                    "version" in dep_data["pkgMeta"]
                    and "update" in dep_data
                    and dep_data["pkgMeta"]["version"] != dep_data["update"]["latest"]
                ):
                    outdated.append(dep)
                elif dep_data.get("incompatible", False):
                    outdated.append(dep)
                else:
                    installed.append(dep)
        # Named dependency not installed
        else:
            missing.append(self.name)

        return installed, missing, outdated

    def install(self):
        return self._exec(["install"])

    def update(self):
        return self._exec(["update"])

    def uninstall(self):
        return self._exec(["uninstall"])


def main():
    arg_spec = dict(
        name=dict(),
        offline=dict(default=False, type="bool"),
        production=dict(default=False, type="bool"),
        path=dict(required=True, type="path"),
        relative_execpath=dict(type="path"),
        state=dict(
            default="present",
            choices=[
                "present",
                "absent",
                "latest",
            ],
        ),
        version=dict(),
    )
    module = AnsibleModule(argument_spec=arg_spec)
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    name = module.params["name"]
    offline = module.params["offline"]
    production = module.params["production"]
    path = module.params["path"]
    relative_execpath = module.params["relative_execpath"]
    state = module.params["state"]
    version = module.params["version"]

    if state == "absent" and not name:
        module.fail_json(msg="uninstalling a package is only available for named packages")

    bower = Bower(
        module,
        name=name,
        offline=offline,
        production=production,
        path=path,
        relative_execpath=relative_execpath,
        version=version,
    )

    changed = False
    if state == "present":
        installed, missing, outdated = bower.list()
        if missing:
            changed = True
            bower.install()
    elif state == "latest":
        installed, missing, outdated = bower.list()
        if missing or outdated:
            changed = True
            bower.update()
    else:  # Absent
        installed, missing, outdated = bower.list()
        if name in installed:
            changed = True
            bower.uninstall()

    module.exit_json(changed=changed)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Tim Cragg (@timcragg)
# Copyright: (c) 2025, Shreyas Srish (@shrsr)
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: aci_l4l7_device
version_added: "2.12.0"
short_description: Manage L4-L7 Devices (vns:LDevVip)
description:
- Manage Layer 4 to Layer 7 (L4-L7) Devices.
options:
  tenant:
    description:
    - The name of an existing tenant.
    type: str
    aliases: [ tenant_name ]
  name:
    description:
    - The name of the L4-L7 device.
    type: str
    aliases: [ device, logical_device, device_name, logical_device_name ]
  context_aware:
    description:
    - Is device Single or Multi context aware.
    - The APIC defaults to C(single) when unset during creation.
    type: str
    choices: [ multi, single ]
  device_type:
    description:
    - The type of the device.
    - The APIC defaults to C(physical) when unset during creation.
    type: str
    choices: [ physical, virtual ]
    aliases: [ dev_type ]
  function_type:
    description:
    - The function type of the device.
    - The APIC defaults to C(go_to) when unset during creation.
    type: str
    choices: [ go_to, go_through, l1, l2 ]
    aliases: [ func_type ]
  managed:
    description:
    - Is the device a managed device.
    - The APIC defaults to true when unset during creation.
    type: bool
  promiscuous_mode:
    description:
    - Enable promiscuous mode.
    - The APIC defaults to false when unset during creation.
    type: bool
    aliases: [ prom_mode ]
  service_type:
    description:
    - The service type running on the device.
    - The APIC defaults to C(others) when unset during creation.
    type: str
    choices: [ adc, fw, others ]
    aliases: [ svc_type ]
  trunking:
    description:
    - Enable trunking.
    - The APIC defaults to false when unset during creation.
    type: bool
  domain:
    description:
    - The domain to bind to the device.
    - The type of domain is controlled by the device_type setting.
    type: str
  active_active_mode:
    description:
    - The active active mode on the device.
    - This is only applicable when C(function_type="l1") or C(function_type="l2").
    - The APIC defaults to false when unset during creation.
    type: bool
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
notes:
- The I(tenant) must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant) modules can be used for this.
seealso:
- module: cisco.aci.aci_tenant
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(vns:LDevVip)
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Tim Cragg (@timcragg)
- Shreyas Srish (@shrsr)
"""

EXAMPLES = r"""
- name: Add a new L4-L7 device
  cisco.aci.aci_l4l7_device:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    name: my_device
    state: present
    domain: phys
    function_type: go_to
    context_aware: single
    managed: false
    device_type: physical
    service_type: adc
    trunking: false
    promiscuous_mode: true
  delegate_to: localhost

- name: Query an L4-L7 device
  cisco.aci.aci_l4l7_device:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    name: my_device
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all L4-L7 devices
  cisco.aci.aci_l4l7_device:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result

- name: Delete an existing L4-L7 device
  cisco.aci.aci_l4l7_device:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    name: my_device
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec
from ansible_collections.cisco.aci.plugins.module_utils.constants import L4L7_FUNC_TYPES_MAPPING


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        tenant=dict(type="str", aliases=["tenant_name"]),
        name=dict(type="str", aliases=["device", "device_name", "logical_device", "logical_device_name"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        context_aware=dict(type="str", choices=["single", "multi"]),
        device_type=dict(type="str", aliases=["dev_type"], choices=["physical", "virtual"]),
        function_type=dict(type="str", aliases=["func_type"], choices=list(L4L7_FUNC_TYPES_MAPPING)),
        managed=dict(type="bool"),
        promiscuous_mode=dict(type="bool", aliases=["prom_mode"]),
        service_type=dict(type="str", aliases=["svc_type"], choices=["adc", "fw", "others"]),
        trunking=dict(type="bool"),
        domain=dict(type="str"),
        active_active_mode=dict(type="bool"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["tenant", "name"]],
            ["state", "present", ["tenant", "name", "device_type", "domain"]],
        ],
    )

    aci = ACIModule(module)

    tenant = module.params.get("tenant")
    state = module.params.get("state")
    name = module.params.get("name")
    context_aware = module.params.get("context_aware")
    device_type = module.params.get("device_type")
    function_type = L4L7_FUNC_TYPES_MAPPING.get(module.params.get("function_type"))
    managed = aci.boolean(module.params.get("managed"))
    promiscuous_mode = aci.boolean(module.params.get("promiscuous_mode"))
    service_type = module.params.get("service_type")
    trunking = aci.boolean(module.params.get("trunking"))
    domain = module.params.get("domain")
    active_active_mode = aci.boolean(module.params.get("active_active_mode"))

    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class="vnsLDevVip",
            aci_rn="lDevVip-{0}".format(name),
            module_object=name,
            target_filter={"name": name},
        ),
        child_classes=["vnsRsALDevToPhysDomP", "vnsCDev", "vnsRsALDevToDomP"],
    )

    aci.get_existing()

    if state == "present":
        child_configs = []
        if device_type == "virtual":
            dom_class = "vnsRsALDevToDomP"
            tdn = "uni/vmmp-VMware/dom-{0}".format(domain)
        else:
            dom_class = "vnsRsALDevToPhysDomP"
            tdn = "uni/phys-{0}".format(domain)
        child_configs.append({dom_class: {"attributes": {"tDn": tdn}}})

        aci.payload(
            aci_class="vnsLDevVip",
            class_config=dict(
                name=name,
                contextAware="{0}-Context".format(context_aware),
                devtype=device_type.upper(),
                funcType=function_type,
                managed=managed,
                promMode=promiscuous_mode,
                svcType=service_type.upper(),
                trunking=trunking,
                activeActive=active_active_mode,
            ),
            child_configs=child_configs,
        )

        aci.get_diff(aci_class="vnsLDevVip")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



# Copyright (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Make coding more python3-ish
from __future__ import annotations

DOCUMENTATION = r"""
name: random_mac
short_description: Generate a random MAC address
description:
  - Generates random networking interfaces MAC addresses for a given prefix.
options:
  _input:
    description: A string prefix to use as a basis for the random MAC generated.
    type: string
    required: true
  seed:
    description:
      - A randomization seed to initialize the process, used to get repeatable results.
      - If no seed is provided, a system random source such as C(/dev/urandom) is used.
    required: false
    type: string
"""

EXAMPLES = r"""
- name: Random MAC given a prefix
  ansible.builtin.debug:
    msg: "{{ '52:54:00' | community.general.random_mac }}"
    # => '52:54:00:ef:1c:03'

- name: With a seed
  ansible.builtin.debug:
    msg: "{{ '52:54:00' | community.general.random_mac(seed=inventory_hostname) }}"
"""

RETURN = r"""
_value:
  description: The generated MAC.
  type: string
"""

import re
from random import Random, SystemRandom

from ansible.errors import AnsibleFilterError


def random_mac(value, seed=None):
    """takes string prefix, and return it completed with random bytes
    to get a complete 6 bytes MAC address"""

    if not isinstance(value, str):
        raise AnsibleFilterError(f"Invalid value type ({type(value)}) for random_mac ({value})")

    value = value.lower()
    mac_items = value.split(":")

    if len(mac_items) > 5:
        raise AnsibleFilterError(f"Invalid value ({value}) for random_mac: 5 colon(:) separated items max")

    err = ""
    for mac in mac_items:
        if not mac:
            err += ",empty item"
            continue
        if not re.match("[a-f0-9]{2}", mac):
            err += f",{mac} not hexa byte"
    err = err.strip(",")

    if err:
        raise AnsibleFilterError(f"Invalid value ({value}) for random_mac: {err}")

    if seed is None:
        r = SystemRandom()
    else:
        r = Random(seed)
    # Generate random int between x1000000000 and xFFFFFFFFFF
    v = r.randint(68719476736, 1099511627775)
    # Select first n chars to complement input prefix
    remain = 2 * (6 - len(mac_items))
    rnd = f"{v:x}"[:remain]
    return value + re.sub(r"(..)", r":\1", rnd)


class FilterModule:
    """Ansible jinja2 filters"""

    def filters(self):
        return {
            "random_mac": random_mac,
        }



#!/usr/bin/python
#
# Copyright (c) 2021 Aparna Patil(@techcon65)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_privatednszonelink

version_added: "1.6.0"

short_description: Create, delete and update Virtual network link for Private DNS zone

description:
    - Creates, deletes, and updates Virtual network links for an existing Azure Private DNS Zone.

options:
    resource_group:
        description:
            - Name of resource group.
        required: true
        type: str
    name:
        description:
            - The name of the virtual network link.
        required: true
        type: str
    zone_name:
        description:
            - The name of the Private DNS zone.
        required: true
        type: str
    registration_enabled:
        description:
            - Is auto-registration of virtual machine records in the virtual network in the Private DNS zone enabled
        default: false
        type: bool
    resolution_policy:
        description:
            - The resolution policy on the virtual network link.
            - Only applicable for virtual network links to privatelink zones, and for A,AAAA,CNAME queries.
            - When set to C(NxDomainRedirect), Azure DNS resolver falls back to public resolution
              if private dns query resolution results in non-existent domain response.
        type: str
        version_added: "3.7.0"
        choices:
            - Default
            - NxDomainRedirect
    virtual_network:
        description:
            - The reference of the virtual network.
        type: str
    state:
        description:
            - Assert the state of the virtual network link. Use C(present) to create or update and C(absent) to delete.
        default: present
        type: str
        choices:
            - absent
            - present

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Aparna Patil (@techcon65)
'''

EXAMPLES = '''
- name: Create a virtual network link
  azure_rm_privatednszonelink:
    resource_group: myResourceGroup
    name: vnetlink1
    zone_name: privatezone.com
    virtual_network: MyAzureVNet
    state: present

- name: Update virtual network link
  azure_rm_privatednszonelink:
    resource_group: myResourceGroup
    name: vnetlink1
    zone_name: privatezone.com
    virtual_network: MyAzureVNet
    registration_enabled: true
    state: present
    tags:
      key1: "value1"

- name: Delete a virtual network link
  azure_rm_privatednszonelink:
    resource_group: myResourceGroup
    name: vnetlink1
    zone_name: privatezone.com
    state: absent
'''

RETURN = '''
state:
    description:
        - Current state of the Virtual network link.
    returned: always
    type: complex
    contains:
        id:
            description:
                - The Virtual network link ID.
            returned: always
            type: str
            sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/
                     Microsoft.Network/privateDnsZones/privatezone.com/virtualNetworkLinks/vnetlink1"
        name:
            description:
                - Virtual network link name.
            returned: always
            type: str
            sample: 'vnetlink1'
        location:
            description:
                - The Azure Region where the resource lives.
            returned: always
            type: str
            sample: global
        etag:
            description:
                - The etag of the virtual network link.
            returned: always
            type: str
            sample: 692c3e92-a618-46fc-aecd-8f888807cd6c
        tags:
            description:
                - Resource tags.
            returned: always
            type: list
            sample: [{"key1": "value1"}]
        virtual_network:
            description:
                - Reference to virtual network.
            returned: always
            type: dict
            sample: {
                "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/
                       providers/Microsoft.Network/virtualNetworks/MyAzureVNet"
            }
        registration_enabled:
            description:
                - The status of auto-registration of virtual machine records in the virtual network in private DNS zone.
            returned: always
            type: bool
            sample: true
        resolution_policy:
            description:
                - The resolution policy on the virtual network link.
            type: str
            returned: always
            sample: Default
        provisioning_state:
            description:
                - The provisioning state of the resource.
            returned: always
            type: str
            sample: Succeeded
        virtual_network_link_state:
            description:
                - The status of the virtual network link.
            returned: always
            type: str
            sample: Completed
'''

from ansible.module_utils.basic import _load_params
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase, \
    format_resource_id

try:
    from azure.core.exceptions import ResourceNotFoundError
    from azure.core.polling import LROPoller
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMVirtualNetworkLink(AzureRMModuleBase):

    def __init__(self):

        _load_params()
        # define user inputs from playbook
        self.module_arg_spec = dict(
            resource_group=dict(type='str', required=True),
            name=dict(type='str', required=True),
            zone_name=dict(type='str', required=True),
            virtual_network=dict(type='str'),
            state=dict(choices=['present', 'absent'], default='present', type='str'),
            registration_enabled=dict(type='bool', default=False),
            resolution_policy=dict(type='str', choices=['Default', 'NxDomainRedirect']),
        )

        required_if = [
            ('state', 'present', ['virtual_network'])
        ]

        self.results = dict(
            changed=False,
            state=dict()
        )

        self.resource_group = None
        self.name = None
        self.zone_name = None
        self.virtual_network = None
        self.registration_enabled = None
        self.state = None
        self.tags = None
        self.resolution_policy = None
        self.log_path = None
        self.log_mode = None

        super(AzureRMVirtualNetworkLink, self).__init__(self.module_arg_spec,
                                                        required_if=required_if,
                                                        supports_tags=True,
                                                        supports_check_mode=True)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        changed = False
        results = dict()
        zone = None
        virtual_network_link_old = None
        virtual_network_link_new = None

        # retrieve resource group to make sure it exists
        self.get_resource_group(self.resource_group)

        if self.virtual_network:
            virtual_network = self.parse_resource_to_dict(self.virtual_network)
            self.virtual_network = format_resource_id(val=virtual_network['name'],
                                                      subscription_id=virtual_network['subscription_id'],
                                                      namespace='Microsoft.Network',
                                                      types='virtualNetworks',
                                                      resource_group=virtual_network['resource_group'])

        self.log('Fetching Private DNS zone {0}'.format(self.zone_name))
        zone = self.private_dns_client.private_zones.get(self.resource_group, self.zone_name)
        if not zone:
            self.fail('The zone {0} does not exist in the resource group {1}'.format(self.zone_name,
                                                                                     self.resource_group))

        try:
            self.log('Fetching Virtual network link {0}'.format(self.name))
            virtual_network_link_old = self.private_dns_client.virtual_network_links.get(self.resource_group,
                                                                                         self.zone_name,
                                                                                         self.name)
            # serialize object into a dictionary
            results = self.vnetlink_to_dict(virtual_network_link_old)
            if self.state == 'present':
                changed = False
                update_tags, results['tags'] = self.update_tags(results['tags'])
                if update_tags:
                    changed = True
                self.tags = results['tags']
                if self.registration_enabled != results['registration_enabled']:
                    changed = True
                    results['registration_enabled'] = self.registration_enabled
                if self.resolution_policy is not None and self.resolution_policy != results.get('resolution_policy'):
                    changed = True
                    results['resolution_policy'] = self.resolution_policy
                else:
                    self.resolution_policy = results.get('resolution_policy')
            elif self.state == 'absent':
                changed = True

        except ResourceNotFoundError:
            if self.state == 'present':
                changed = True
            else:
                changed = False

        self.results['changed'] = changed
        self.results['state'] = results

        if self.check_mode:
            return self.results

        if changed:
            if self.state == 'present':
                # create or update Virtual network link
                virtual_network_link_new = \
                    self.private_dns_models.VirtualNetworkLink(location='global',
                                                               resolution_policy=self.resolution_policy,
                                                               registration_enabled=self.registration_enabled)
                if self.virtual_network:
                    virtual_network_link_new.virtual_network = \
                        self.network_models.VirtualNetwork(id=self.virtual_network)
                if self.tags:
                    virtual_network_link_new.tags = self.tags
                self.results['state'] = self.create_or_update_network_link(virtual_network_link_new)

            elif self.state == 'absent':
                # delete virtual network link
                self.delete_network_link()
                self.results['state'] = 'Deleted'

        return self.results

    def create_or_update_network_link(self, virtual_network_link):
        try:
            # create the virtual network link
            response = \
                self.private_dns_client.virtual_network_links.begin_create_or_update(resource_group_name=self.resource_group,
                                                                                     private_zone_name=self.zone_name,
                                                                                     virtual_network_link_name=self.name,
                                                                                     parameters=virtual_network_link)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as exc:
            self.fail("Error creating or updating virtual network link {0} - {1}".format(self.name, str(exc)))
        return self.vnetlink_to_dict(response)

    def delete_network_link(self):
        try:
            # delete the virtual network link
            response = self.private_dns_client.virtual_network_links.begin_delete(resource_group_name=self.resource_group,
                                                                                  private_zone_name=self.zone_name,
                                                                                  virtual_network_link_name=self.name)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as exc:
            self.fail("Error deleting virtual network link {0} - {1}".format(self.name, str(exc)))
        return response

    def vnetlink_to_dict(self, virtualnetworklink):
        result = virtualnetworklink.as_dict()
        result['tags'] = virtualnetworklink.tags
        return result


def main():
    AzureRMVirtualNetworkLink()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Bill Peck, <@p3ck>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


class ModuleDocFragment(object):

    # Azure doc fragment
    DOCUMENTATION = r'''
options:
    azure_active_directory_application_id_or_uri:
        description:
            - The Azure Active Directory Application Id or Uri to get the access token
              that will be included as the bearer token in delivery requests. Applicable
              only for webhook as a destination.
        required: false
        type: str
    azure_active_directory_tenant_id:
        description:
            - The Azure Active Directory Tenant Id to get the access token that will be
              included as the bearer token in delivery requests. Applicable only for
              webhook as a destination.
        required: false
        type: str
    deadletter_endpoint:
        description:
            - The Azure resource ID of an Azure Storage blob container destination where
              EventGrid should deadletter undeliverable events for this event subscription.
        required: false
        type: str
    delivery_attribute_mapping:
        description:
            - Add delivery attribute mapping to send additional information via HTTP headers
              when delivering events. This attribute is valid for all destination types except
              StorageQueue.
            - Format attribute-name attribute-type attribute-value [attribute-is-secret]
            - somename dynamic somevalue true
            - somename2 static somevalue
        type: list
        elements: str
    endpoint:
        description:
            - Endpoint where EventGrid should deliver events matching this event subscription.
              For webhook endpoint type, this should be the corresponding webhook URL. For
              other endpoint types, this should be the Azure resource identifier of the endpoint.
              It is expected that the destination endpoint to be already created and available
              for use before executing any Event Grid command.
        required: false
        type: str
    endpoint_type:
        description:
            - The type of the destination endpoint.
        type: str
        choices:
            - azurefunction
            - eventhub
            - hybridconnection
            - servicebusqueue
            - servicebustopic
            - storagequeue
            - webhook
        default: webhook
    event_delivery_schema:
        description:
            - The schema in which events should be delivered for this event subscription.
              By default, events will be delivered in the same schema in which they are
              published (based on the corresponding topic's input schema).
        type: str
        choices:
            - CloudEventSchemaV1_0
            - CustomInputSchema
            - EventGridSchema
    event_ttl:
        description:
            - Event time to live (in minutes). Must be a number between 1 and 1440.
        type: int
        default: 1440
    expiration_date:
        description:
            - Date or datetime (in UTC, e.g. '2018-11-30T11:59:59+00:00' or '2018-11-30')
              after which the event subscription would expire. By default, there is no expiration
              for the event subscription.
        type: str
    labels:
        description:
            - A list of labels to associate with this event subscription.
        type: list
        elements: str
    max_delivery_attempts:
        description:
            - Maximum number of delivery attempts. Must be a number between 1 and 30.
        type: int
        default: 30
    max_events_per_batch:
        description:
            - Maximum number of events in a batch. Must be a number between 1 and 5000.
        type: int
    preferred_batch_size_in_kilobytes:
        description:
            - Preferred batch size in kilobytes. Must be a number between 1 and 1024.
        type: int
    storage_queue_msg_ttl:
        description:
            - Storage queue message time to live in seconds.
        type: int
    advanced_filter:
        description:
            - An advanced filter enables filtering of events based on a specific event property.
            - KEY[.INNERKEY] FILTEROPERATOR VALUE [VALUE ...]
            - Examples
            - data.Color StringIn Blue Red Orange Yellow
            - data.Color StringNotIn Blue Red Orange Yellow
            - subject StringContains Blue Red
            - subject StringNotContains Blue Red
            - subject StringBeginsWith Blue Red
            - subject StringNotBeginsWith Blue Red
            - subject StringEndsWith img png jpg
            - subject StringNotEndsWith img png jpg
            - data.property1 NumberIn 5 10 20
            - data.property1 NumberInRange 5,10 20,30 40,50
            - data.property2 NumberNotIn 100 200 300
            - data.property2 NumberNotInRange 100,110 200,210 300,310
            - data.property3 NumberLessThan 100
            - data.property2 NumberLessThanOrEquals 100
            - data.property3 NumberGreaterThan 100
            - data.property2 NumberGreaterThanOrEquals 100
            - data.property3 BoolEquals true
            - data.property3 IsNullOrUndefined
            - data.property3 IsNotNull
        type: list
        elements: str
    enable_advanced_filtering_on_arrays:
        description:
            - Allows advanced filters to be evaluated against an array of values
              instead of expecting a singular value.
        type: bool
    included_event_types:
        description:
            - A list of event types (e.g., Microsoft.Storage.BlobCreated and
              Microsoft.Storage.BlobDeleted). In order to subscribe to all default event types,
              do not specify any value for this argument. For event grid topics, event types are
              customer defined. For Azure events, e.g., Storage Accounts, IoT Hub, etc.,
              you can query their event types using this CLI command
              'az eventgrid topic-type list-event-types'.
        type: list
        elements: str
    subject_begins_with:
        description:
            - An optional string to filter events for an event subscription based on a prefix.
              Wildcard characters are not supported.
        type: str
    subject_case_sensitive:
        description:
            - Specify to indicate whether the subject fields should be compared in a case sensitive
              manner. True if flag present.
        type: bool
    subject_ends_with:
        description:
            - An optional string to filter events for an event subscription based on a suffix.
              Wildcard characters are not supported.
        type: str
'''
